from typing import Any, Callable, Dict, Type, Union, List, Tuple
from typing_extensions import Self
import pandas as pd
import logging

from .models import ProphetModel
from ..utility.check_utils import check_not_isinstance, check_not_in_iterable
from ..utility.resources import get_module_and_function, log_exception, logger

@log_exception(logger)
class ExperimentClass():
    def __init__(self,  model: str, params : Dict[str, Union[Dict[str, Union[int, float, str]], List[str]]] , config_path : str):
        """ 
        Init of the ExperimentClass
        The class instantiates the wanted model based on the <model> parameter

        Parameters
        ----------
        model : str
            Parameter identifying model to use
        params : Dict[str, Union[Dict[str, Union[int, float, str]], List[str]]]         
            Dictionary representing the model parameters
        config_path : str
            Path in which resides the configuration files 
            
        """
        self.model_name = model.lower()
        self.config_path = config_path

        if self.model_name == 'prophet':
            log_del = logging.getLogger('cmdstanpy')
            log_del.addHandler(logging.NullHandler())
            log_del.propagate = False
            log_del.setLevel(logging.CRITICAL)

            self.Model = ProphetModel(params=params, config_path=self.config_path)
        else:
            raise Exception()


    def check_consistency(self, X: pd.DataFrame):
        """
        Check if the dataframe is consistent for the next operations

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe to check
        """

        check_not_isinstance(obj=X, data_type=pd.DataFrame, func=get_module_and_function())
        check_not_in_iterable(obj='unique_id', iterable=X.reset_index().columns, func=get_module_and_function())
        check_not_in_iterable(obj='ds', iterable=X.columns, func=get_module_and_function())
        check_not_in_iterable(obj='y', iterable=X.columns, func=get_module_and_function())


    def fit(self, X_train: pd.DataFrame, X_val: pd.DataFrame, y=None) -> Self:
        """
        Fit method of ExperimentClass

        Parameters
        ----------
        X_train : pd.DataFrame
            Dataframe with the train observations
        X_val : pd.DataFrame
            Dataframe with the validation observations
        y : _type_, optional
            Target_value, by default None
            This parameter is not used
        Returns
        -------
        Self
        """

        for x in [X_train,X_val]:
            self.check_consistency(X=x)
        
        self.Model.fit(X_train=X_train, X_val=X_val)

    
    def refit(self, X: pd.DataFrame):
        """
        Refit method of ExperimentClass

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe on which the function refit the data
        """
        
        self.check_consistency(X=X)

        self.Model.refit(X=X)
       

    def predict(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """
        Predict method which generates predictions given an input dataframe

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe from which generate the predictions
        y : _type_, optional
            Target_value, by default None
            This parameter is not used
        Returns
        -------
        pd.DataFrame
            Output dataframe
        """

        self.check_consistency(X=X)

        output = self.Model.predict(X=X)

        return output


    def fit_predict(self, X_train: pd.DataFrame, X_val: pd.DataFrame, X_test: pd.DataFrame) -> pd.DataFrame:
        """
        fit_predict method of ExperimentClass

        Parameters
        ----------
        X_train : pd.DataFrame
            Dataframe with the train observations
        X_val : pd.DataFrame
            Dataframe with the validation observations
        X_test : pd.DataFrame
            Dataframe with the test observations

        Returns
        -------
        pd.DataFrame
            _description_
        """

        for x in [X_train, X_val, X_test]:
            self.check_consistency(X=x)
        
        output = self.Model.fit(X_train=X_train, X_val=X_val).predict(X=X_test)

        return output
    

    def create_hyperparameters_table(self, hyperparameters: dict[str, Any]) -> dict[str, Any]:
        """
        The function creates the hyperparameters table ready to be logged on mlflow

        Parameters
        ----------
        hyperparameters : dict[str, Any]
            Hyperparameters from the configuration file

        Returns
        -------
        dict[str, Any]
            Dictionary of hyperparameters to log
        """
        def get_params_to_log(model: Type[Callable], hyperparameters: list[str]) -> dict[str, Any]:
            """
            The function takes the parameters to log from the hyperparameters of the configuration file

            Parameters
            ----------
            model : Type[Callable]
                Model from which the parameters are obtained
            hyperparameters : list[str]
                Hyperparameters needed to obtain their value from the model

            Returns
            -------
            dict[str, Any]
                Dictionary of hyperparameters to log
            """
            return {hyper: getattr(model, hyper) for hyper in hyperparameters}

        return {id_pred.replace('/', '_') : get_params_to_log(model=model, hyperparameters=list(hyperparameters.keys())) if model else None 
                for id_pred, model in self.Model.models_dict.items()}


    def score(self, X_true: pd.DataFrame, X_pred: pd.DataFrame, y=None, full_scoring=True) -> Tuple[pd.Series, Union[pd.DataFrame, None]]:
        """
        Creates the scores given two input dataframes (predictions and real values)
        using the EvaluateModel class

        Parameters
        ----------
        X_true : pd.DataFrame
            Dataframe containing the real values
        X_pred : pd.DataFrame
            Dataframe containing the predictions
        y : _type_, optional
            Target_value, by default None
            This parameter is not used
        full_scoring : bool, optional
            Boolean value to switch to True when we need also the full scores for all the models, by default True

        Returns
        -------
        pd.DataFrame
            Tuple[pd.Series, Union[pd.DataFrame, None]]
            Tuple representing the scores and eventually the full scoring
        """
        self.check_consistency(X=X_true)
        check_not_in_iterable(obj='Id_Pred', iterable=X_pred.columns, func=get_module_and_function())
        check_not_in_iterable(obj='Pred_Mean', iterable=X_pred.columns, func=get_module_and_function())
        check_not_in_iterable(obj='Pi_Upper_95', iterable=X_pred.columns, func=get_module_and_function())
        check_not_in_iterable(obj='Pi_Lower_95', iterable=X_pred.columns, func=get_module_and_function())

        logger.info(message = f"START Evaluating Model.")
        scores = self.Model.score(X_true=X_true, X_pred=X_pred)

        scores_aggr = scores.drop(columns='Id_Pred')
        scores_aggr = scores_aggr.mean()
        scores_aggr.name = 'Metrics'

        logger.info(message = f"DONE Evaluating Model.")

        if full_scoring:
            return scores_aggr, scores
        else:
            return scores_aggr, None


import pandas as pd
import pytest
import numpy as np
from test_tc.datapreparation.utils import (Normalizer, ReplacerNA, Detrender, Deseasoner, Differencer, 
                                           OutlierRemover, Smoother, Differentiator, LogTransformer,
                                           map_columns, get_hierarchical_df, get_hierarchical_info, missing_data_imputation)
from statsmodels.tsa.seasonal import seasonal_decompose
from test_tc.datapreparation.utils import datepart,get_hierarchy_column,divide_by_father,get_lagged_column

### Normalizer ###
@pytest.fixture
def example_data():
    # Example data for testing
    data = pd.DataFrame({'volumes': [10, 20, 30, 40, 50]})
    return data

def test_fit_normalizer(example_data):
    # Create an instance of the Normalizer class
    normalizer = Normalizer()

    # Call the fit method with the example data
    fitted_normalizer = normalizer.fit(example_data)

    # Verify that the output is the same instance of normalizer
    assert fitted_normalizer is normalizer

    # Verify that the min and max values are calculated correctly
    assert normalizer.min == 10
    assert normalizer.max == 50

def test_transform_normalizer(example_data):
    # Create an instance of the Normalizer class
    normalizer = Normalizer()

    # Call the fit method with the example data
    normalizer.fit(example_data)

    # Call the transform method with the example data
    transformed_data = normalizer.transform(example_data)

    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the DataFrame has the correct columns
    assert transformed_data.columns.tolist() == ['volumes']

    # Verify that the data is correctly normalized
    assert transformed_data['volumes'].min() == 0
    assert transformed_data['volumes'].max() == 1

def test_inverse_transform_normalizer(example_data):
    normalizer = Normalizer()

    # Call the fit method with the example data
    normalizer.fit(example_data)

    # Call the transform method with the example data
    transformed_data = normalizer.transform(example_data)

    data = normalizer.inverse_transform(transformed_data)
    assert np.min(data) == normalizer.min
    assert np.max(data) == normalizer.max

### ReplacerNA ###
#transform function
@pytest.fixture
def example_data_transform():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, None, 30, None, 50]})
    return data

def test_fit_replace_na(example_data_transform):
    # Create an instance of the ReplacerNA class with method="mean"
    replacer = ReplacerNA(method="mean")

    # Call the fit method with the example data
    fitted_replacer = replacer.fit(example_data_transform)

    # Verify that the output is the same instance of replacer
    assert fitted_replacer is replacer

    # Verify that the value and method_for_df are calculated correctly
    assert replacer.value == 30.0
    assert replacer.method_for_df is None

def test_transform_replace_na(example_data_transform):
    # Create an instance of the ReplacerNA class with method="zero"
    replacer = ReplacerNA(method="zero")

    # Call the fit method with the example data
    replacer.fit(example_data_transform)

    # Call the transform method with the example data
    transformed_data = replacer.transform(example_data_transform)
    print(transformed_data)
    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the transformed data has missing values replaced correctly
    expected_data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                                  'volumes': [10.0, 0.0, 30.0, 0.0, 50.0]})
    pd.testing.assert_frame_equal(transformed_data, expected_data)

def test_transform__replace_na_ffill(example_data_transform):
    # Create an instance of the ReplacerNA class with method="ffill"
    replacer = ReplacerNA(method="ffill")

    # Call the fit method with the example data
    replacer.fit(example_data_transform)

    # Call the transform method with the example data
    transformed_data = replacer.transform(example_data_transform)

    # Verify that the transformed data has missing values replaced correctly using forward fill
    expected_data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                                  'volumes': [10.0, 10.0, 30.0, 30.0, 50.0]})
    pd.testing.assert_frame_equal(transformed_data, expected_data)


### Detrender ###
#trend function
@pytest.fixture
def example_data_trend():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, 20, 30, 40, 50]})
    return data

def test_fit_detrender(example_data_trend):
    # Create an instance of the Detrender class with period=2
    detrender = Detrender(period=2)

    # Call the fit method with the example data
    fitted_detrender = detrender.fit(example_data_trend)

    # Verify that the output is the same instance of detrender
    assert fitted_detrender is detrender

    # Verify that the trend is calculated correctly
    expected_trend = pd.Series([10.0, 20.0, 30.0, 40.0, 50.0], name='trend')
    pd.testing.assert_series_equal(detrender.trend, expected_trend)

def test_transform_detrender(example_data_trend):
    # Create an instance of the Detrender class with period=2
    detrender = Detrender(period=2)

    # Call the fit method with the example data
    detrender.fit(example_data_trend)

    # Call the transform method with the example data
    transformed_data = detrender.transform(example_data_trend)

    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the transformed data is detrended correctly
    expected_trend = seasonal_decompose(example_data_trend['volumes'], model='additive', period=2, extrapolate_trend='freq').trend
    expected_data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                                  'volumes': example_data_trend['volumes'] - expected_trend})
    pd.testing.assert_frame_equal(transformed_data, expected_data)


### Deseasoner ###
#season function
@pytest.fixture
def example_data_season():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, 20, 30, 40, 50]})
    return data

def test_transform_deseasoner(example_data_season):
    # Create an instance of the Deseasoner class with period=2
    deseasoner = Deseasoner(period=2)

    # Call the fit method with the example data
    deseasoner.fit(example_data_season)

    # Call the transform method with the example data
    transformed_data = deseasoner.transform(example_data_season)

    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the transformed data is deseasonalized correctly
    expected_seasonal = seasonal_decompose(example_data_season['volumes'], model='additive', period=2, extrapolate_trend='freq').seasonal
    expected_data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                                  'volumes': example_data_season['volumes'] - expected_seasonal})
    pd.testing.assert_frame_equal(transformed_data, expected_data)


### Differencer ###
#difference function
@pytest.fixture
def example_data_difference():
    # Example data for testing
    data = pd.DataFrame({'timestamp': [1, 2, 3, 4, 5],
                         'volumes': [10, 20, 30, 40, 50]})
    return data

def test_transform_differencer(example_data_difference):
    # Create an instance of the Differencer class with lag=1
    differencer = Differencer(lag=1)

    # Call the fit method with the example data
    differencer.fit(example_data_difference)

    # Call the transform method with the example data
    transformed_data = differencer.transform(example_data_difference)

    # Verify that the output is a DataFrame
    assert isinstance(transformed_data, pd.DataFrame)

    # Verify that the transformed data is differenced correctly
    expected_data = pd.DataFrame({'timestamp': [2, 3, 4, 5],
                                  'volumes': [10, 10, 10, 10]})
    pd.testing.assert_frame_equal(transformed_data, expected_data)

### OutlierRemover ###
@pytest.fixture
def sample_data_outlier():
    return np.array([10, 15, 20, 25, 30, 100, 200, 300])

def test_outlier_removal(sample_data_outlier):
    lower_threshold_percentile = 5
    upper_threshold_percentile = 95
    outlier_remover = OutlierRemover(lower_threshold_percentile, upper_threshold_percentile)

    # Sample data with outliers
    data = sample_data_outlier

    # Fit the outlier remover on the data
    outlier_remover.fit(data)

    # Transform the data by capping the outliers
    transformed_data = outlier_remover.transform(data)

    # Check if the transformed data has no outliers
    assert np.all(transformed_data >= 10)  # Values are not below the lower threshold
    assert np.all(transformed_data <= 300)  # Values are not above the upper threshold
    assert np.all(transformed_data == data)  # The original non-outlier data remains unchanged


### Smoother ###
@pytest.fixture
def sample_data_smoother():
    # Sample data for testing
    data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    return pd.Series(data)

def test_smoother_transform(sample_data_smoother):
    window_size = 3
    smoother = Smoother(window_size)

    # Sample data for testing
    data = sample_data_smoother

    # Fit the smoother (not necessary in this case)
    smoother.fit(data)

    # Transform the data using the moving average window
    transformed_data = smoother.transform(data)

    # Expected smoothed data using the given window size (calculated manually)
    expected_smoothed_data = [1.0, 1.5, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]

    # Check if the transformed data matches the expected smoothed data
    assert np.allclose(transformed_data, expected_smoothed_data)

### Differenciator ###
@pytest.fixture
def sample_data_diff():
    # Sample data for testing
    data = np.array([1, 4, 6, 9, 12, 15])
    return pd.Series(data)

def test_differentiator_transform(sample_data_diff):
    order = 2
    differentiator = Differentiator(order)

    # Sample data for testing
    data = sample_data_diff

    # Fit the differentiator (not necessary in this case)
    differentiator.fit(data)

    # Transform the data by taking differences between consecutive values (twice in this case)
    transformed_data = differentiator.transform(data)

    # Expected differentiated data using the given order (calculated manually)
    expected_transformed_data = [0., 0., -1., 1., 0., 0.]

    # Check if the transformed data matches the expected differentiated data
    assert np.allclose(transformed_data, expected_transformed_data)


### LogTransformer ###
@pytest.fixture
def sample_data_log():
    # Sample data for testing
    data = np.array([1, 2, 3, 4, 5])
    return data

def test_log_transformer_transform(sample_data_log):
    transformer = LogTransformer()

    # Sample data for testing
    data = sample_data_log

    # Fit the transformer (not necessary in this case)
    transformer.fit(data)

    # Transform the data using the logarithmic transformation
    transformed_data = transformer.transform(data)

    # Expected transformed data (calculated manually)
    expected_transformed_data = [0.69314718, 1.09861229, 1.38629436, 1.60943791, 1.79175947]

    # Check if the transformed data matches the expected transformed data
    assert np.allclose(transformed_data, expected_transformed_data)

def test_log_transformer_inverse_transform(sample_data_log):
    transformer = LogTransformer()

    # Sample transformed data for testing
    transformed_data = [0.69314718, 1.09861229, 1.38629436, 1.60943791, 1.79175947]

    # Fit the transformer (not necessary in this case)
    transformer.fit(sample_data_log)

    # Inverse transform the data using the inverse logarithmic transformation
    reconstructed_data = transformer.inverse_transform(transformed_data)

    # Check if the reconstructed data matches the original data
    assert np.allclose(reconstructed_data, sample_data_log)



### Test Other Utils ###

## test map_columns ##
def sample_data():
    # Sample data for testing
    data = {
        "column_a": [1, 2, 3],
        "column_b": [4, 5, 6],
        "column_c": [7, 8, 9]
    }
    return pd.DataFrame(data)

def test_map_columns():
    hierarchy = {
        "level_1": "column_a",
        "level_2": "column_b",
        "level_3": "column_c"
    }
    conversion = {
        "level_1": None,
        "level_2": None,
        "level_3": "{7: 'seven', 8: 'eight', 9: 'nine'}"
    }

    # Sample data for testing
    df = sample_data()

    # Map the columns using the given hierarchy and conversion dictionaries
    mapped_df = map_columns(hierarchy, df, conversion)

    # Check if the mapped DataFrame has the correct column names
    assert "column_a" in mapped_df.columns
    assert "column_b" in mapped_df.columns
    assert "column_c" in mapped_df.columns

    # Check if the columns were correctly mapped based on the conversion dictionary
    assert mapped_df["column_c"].tolist() == ['seven', 'eight', 'nine']
   

## test get_hierarchical_df ##
def sample_data_hier():
    # Sample data for testing
    data = {
        "date_col": pd.date_range("2023-07-01", periods=5, freq="D"),
        "zone": ["North", "North", "South", "South", "West"],
        "city": ["Milan", "Turin", "Rome", "Naples", "Florence"]
    }
    return pd.DataFrame(data)

import warnings
def test_get_hierarchical_df():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        hierarchy = {
            "level1": "zone",
            "level2": "city"
        }
        time_granularity = "D"
        date_col = "date_col"

        # Sample data for testing
        df = sample_data_hier()

        # Generate the hierarchical DataFrame using the given hierarchy and time_granularity
        hierarchical_df = get_hierarchical_df(df, hierarchy, time_granularity, date_col)

    # Check if the hierarchical DataFrame has the expected columns
    expected_columns = ["ds", "y", "level0", "level1", "level2"]
    for col in hierarchical_df.columns:
        assert col in expected_columns

    # Check if level0 is set as "Italia" for all observations
    assert hierarchical_df["level0"].unique() == ["Italia"]

    # Check if y is set as 1 for all observations
    assert hierarchical_df["y"].unique() == [1]

    # Check if the time_granularity is correctly applied
    assert pd.infer_freq(hierarchical_df["ds"]) == time_granularity


## test get_hierarchical_INFO ##
def test_get_hierarchical_info():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        hierarchy = {
            "level1": "zone",
            "level2": "city"
        }
        time_granularity = "D"
        date_col = "date_col"

        # Sample data for testing
        df = sample_data_hier()

        # Generate the hierarchical DataFrame using the given hierarchy and time_granularity
        hierarchical_df = get_hierarchical_df(df, hierarchy, time_granularity, date_col)
        Y_df, S_df, tags = get_hierarchical_info(hierarchical_df)

        
        # Check the tags consistency
        reference_tags = {'level0': np.array(['Italia'], dtype=object), 'level0/level1': np.array(['Italia/North', 'Italia/South', 'Italia/West'], dtype=object), 
         'level0/level1/level2': np.array(['Italia/North/Milan', 'Italia/North/Turin', 'Italia/South/Naples', 'Italia/South/Rome', 'Italia/West/Florence'])}
        for k,v in tags.items():
            assert np.all(v == reference_tags[k])

        # Check is S_df is a binary matrix
        S_array = S_df.to_numpy()
        assert np.all(np.logical_or(S_array == 0, S_array == 1))

        # Check Y_df consistency
        # the number of rows must be given by len(date_col) * (num_possible_hierarchies + 1 (Italy))
        assert Y_df.shape[0] == len(hierarchical_df['ds']) * S_df.shape[0]


## test missing_data_imputation ##
def sample_data_missing():
    # Sample data for testing
    data = {
        "date_col": pd.date_range("2023-07-01", periods=5, freq="D"),
        "y": [1, None, 3, 4, None],
        "other_column": [5, 6, None, 8, 9],
    }
    return pd.DataFrame(data)

def test_missing_data_imputation_mean():
    # Test missing data imputation with strategy "mean"

    strategy = "mean"

    # Sample data for testing
    df = sample_data_missing()

    # Impute missing values using the mean strategy
    imputed_df = missing_data_imputation(strategy, df)
    # Check if missing values in the 'y' column are replaced with the mean
    assert np.allclose(np.array(imputed_df["y"],dtype='float'), [1, 2.67, 3, 4, 2.67], rtol = 0.01)

def test_missing_data_imputation_spline():
    # Test missing data imputation with strategy "spline"

    strategy = {"interpolation": "spline", "order": 2}

    # Sample data for testing
    df = sample_data_missing()

    # Impute missing values using the spline strategy
    imputed_df = missing_data_imputation(strategy, df)
    
    # Check if the 'y' column is correctly imputed using the spline method
    assert np.allclose(np.array(imputed_df["y"],dtype='float'), [1, 2, 3, 4, 5], rtol = 0.01)

######## Test for xgboost pre-processing ##########################
@pytest.fixture
def datepart_data():

    dates = pd.Series(['2022-01-01 12:00:00', '2022-01-02 18:00:00'])

    dates_true = pd.DataFrame([[1,5,1,2022,1,1,True,True],[2,6,2,2022,1,1,False,True]],
                               columns=["dayofmonth","dayofweek","dayofyear","year","month","quarter","holiday","weekend"],
                               index=['2022-01-01 12:00:00', '2022-01-02 18:00:00'],dtype=np.int32)
    dates_true = dates_true.astype(dtype={"holiday":bool,"weekend":bool})

    return [(dates,dates_true)]

def test_datepart(datepart_data):

    for dates,df in datepart_data:
        assert datepart(dates).equals(df)


@pytest.mark.parametrize('n, length', [(1,1),(2,1),(3,2)])
def test_get_hierarchy_column(n,length):

    df = pd.DataFrame([["A",2022,10],["A/B",2022,5],["A/C",2022,5],["A/B/B1",2022,2],["A/B/B2",2022,3],["A/C/C1",2022,5],
                   ["A/B/B1/B1.1",2022,1],["A/B/B1/B1.2",2022,1],["A/B/B2/B2.1",2022,1],["A/B/B2/B2.2",2022,1],["A/B/B2/B2.3",2022,1]],
                  columns = ["Id","date","pred"])
    
    # livello con father == livello precedente con figli
    lev1 = get_hierarchy_column(df,"date","pred","Id",n+1,sons=False,father=True,type_="COL")
    lev2 = get_hierarchy_column(df,"date","pred","Id",n,sons=True,father=False,type_="COL")
    for i,j in zip(lev1,lev2):
        assert i.equals(j)

    # lunghezza lista (se primo livello -- len sempre 1)
    assert len(get_hierarchy_column(df,"date","pred","Id",n,sons=False,father=True,type_="COL")) == length
    assert len(get_hierarchy_column(df,"date","pred","Id",1,sons=True,father=True,type_="COL")) == 1

    data = get_hierarchy_column(df,"date","pred","Id",n,sons=False,father=True,type_="COL")
    data_real = []

    check_df = []
    for i,j in zip(data,data_real):
        check_df += [i.equals(j)]
    
    assert all(check_df)


@pytest.fixture
def divide_by_father_data():
    
    data = {'A/B': [3, 3, 3], 'A/C': [4, 5, 6], 'A': [7, 8, 9]}
    df = pd.DataFrame(data)

    df_real = pd.DataFrame([[0.428571,0.571429,7],[0.375000,0.625000,8],[0.333333,0.666667,9]],
                           columns=["A/B","A/C","A"])
    
    return [(df,df_real)]
    
def test_divide_by_father(divide_by_father_data):

    for df,df_real in divide_by_father_data:
        df_div = divide_by_father(df)
        pd.testing.assert_frame_equal(df_div,df_real)
        pd.testing.assert_series_equal(df_div.iloc[:,:-1].sum(axis=1),pd.Series([1. for i in range(df.shape[0])]))

@pytest.fixture
def get_lagged_column_data():
    
    data = {'A': [1, 2, 3, 4], 'B': [5, 6, 7, 8]}
    df = pd.DataFrame(data)

    df_real = pd.DataFrame([[np.nan,np.nan],[1,np.nan],[2,1],[3,2]],
                           columns=["Lag1_A","Lag2_A"])
    
    return [(df,df_real)]
    
def test_get_lagged_column(get_lagged_column_data):

    for df,df_real in get_lagged_column_data:
        df_lag = get_lagged_column(df,[1,2],"A")
        pd.testing.assert_frame_equal(df_lag,df_real)









from test_tc.datapreparation.preprocessing import PreprocessingClass

import sys
import pytest
import pandas as pd


def sample_data_hier():
    data = {
        "date_col": pd.date_range("2023-07-01", periods=5, freq="D"),
        "zone": ["North", "North", "South", "South", "West"],
        "city": ["Milan", "Turin", "Rome", "Naples", "Florence"],
        'target_col': ['a','a','b', 'c', 'd']
    }
    return pd.DataFrame(data)

import warnings
def test_PreprocessingClass_TCPreprocess():
    # Test PreprocessingClass with TCPreprocess
    usecase = "teleconsulto"
    model = "prophet"

    hierarchy = {
        "level_1": "zone",
        "level_2": "city"
    }

    conversion = {
        "level_1": None,
        "level_2": None
    }

    target_col = "target_col"
    time_granularity = "D"
    date_col = "date_col"
    missing_data_strategy = "mean"

    # Sample data for testing
    df = sample_data_hier()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        preprocessor = PreprocessingClass(
            usecase,
            model,
            hierarchy=hierarchy,
            conversion=conversion,
            target_col=target_col,
            time_granularity=time_granularity,
            date_col=date_col,
            missing_data_strategy=missing_data_strategy
        )

        # Test transform method
        # The checks on these results are already performed in other tests
        Y_df, S_df, tags = preprocessor.transform(df)
        assert type(Y_df) == pd.DataFrame
        assert type(S_df) == pd.DataFrame
        assert type(tags) == dict
 

def test_PreprocessingClass_TVPreprocess():
    # Test PreprocessingClass with TVPreprocess
    # TV doesn't aplly anything, it should only apply Prophet processing

    usecase = "televisita"
    model = "prophet"
    hierarchy = {
        "level_1": "zone",
        "level_2": "city"
    }

    conversion = {
        "level_1": None,
        "level_2": None
    }

    time_granularity = "D"
    date_col = "date_col"
    missing_data_strategy = "mean"
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        preprocessor = PreprocessingClass(
                usecase,
                model,
                hierarchy=hierarchy,
                conversion=conversion,
                time_granularity=time_granularity,
                date_col=date_col,
                missing_data_strategy=missing_data_strategy
            )

        df = sample_data_hier()

        # TVPreprocess transform should return the DataFrame unchanged
        Y_df, S_df, tags = preprocessor.transform(df)
    assert type(Y_df) == pd.DataFrame
    assert type(S_df) == pd.DataFrame
    assert type(tags) == dict


class ExitException(Exception):
    pass

def my_exit(status=None):
    raise ExitException(f"SystemExit with status: {status}")

def test_PreprocessingClass_InvalidUsecase(monkeypatch):
    # Test PreprocessingClass with an invalid usecase

    usecase = "invalid_usecase"
    model = "prophet"

    # Patch sys.exit with a custom function
    monkeypatch.setattr(sys, "exit", my_exit)

    with pytest.raises(Exception):
        PreprocessingClass(usecase, model)

def test_PreprocessingClass_InvalidModel(monkeypatch):
    # Test PreprocessingClass with an invalid model

    usecase = "teleconsulto"
    model = "invalid_model"

    # Patch sys.exit with a custom function
    monkeypatch.setattr(sys, "exit", my_exit)
    with pytest.raises(Exception):
        PreprocessingClass(usecase, model)












import os
import mlflow
import pandas as pd
import json
import numpy as np


from ..utility.resources import log_exception, get_module_and_function, logger
from ..utility.check_utils import check_not_in_iterable, check_not_isinstance
from ..algorithms.models import HierarchicalReconcileModel


@log_exception(logger)
class JobClass():
    def __init__(self, model_name: str, time_granularity: str, mlflow_runs: pd.DataFrame, apply_hierarchical: bool):
        """
        Init of JobClass

        Parameters
        ----------
        model_name : str
            Represents the name of the model which predictions must be reconciled
        time_granularity : str
            Time aggregation on which the date column is converted
        mlflow_runs : pd.DataFrame
            The selected runs loaded from mlflow
        apply_hierarchical : bool
            Boolean value which specify if the hierarchical reconciliation is applied
        """
        
        self.model_name = model_name
        self.runs = mlflow_runs
        self.apply_hierarchical = apply_hierarchical
        self.time_granularity = time_granularity


    def find_best_run(self, decision_function: str):
        """
        Returns the best model from a set of experiments given two metrics

        Parameters
        ----------
        decision_function : str
            Decision function to minimize in order to find the best parameters.

            The Decision function can use ONLY the evaluation parameters, beware that if an evaluation parameter
            has a space " " in its name, inside the evaluation it MUST be written as a "_", for example if using half
            Percentage Coverage and half RMSE the function will be 'percentage_coverage*0.5 + rmse*0.5'
        """
        logger.info(message = "Parsing and Selecting MLFlow runs.")
        slicer =[]
        for idx, row in self.runs.iterrows():
            try:
                num_models = len(json.loads(row['tags.mlflow.log-model.history']))
                if self.apply_hierarchical and num_models == 2:
                    slicer.append(idx)
                elif not self.apply_hierarchical and num_models == 1:
                    slicer.append(idx)
            except:
                logger.warning(message = f'An Error has been encountered while reading run {row.run_id} for experiment {row.experiment_id}. Skipping the run.')
        
        if len(slicer) != 0:
            sliced_runs = self.runs.iloc[slicer]
        else:
            raise Exception('Trying to get a best run, but none available are found in the runs history. Check the if the model you are trying to use to predict has ever been executed before.')  

        logger.info(message = "Evaluating Best MLFlow run.")
        min = np.inf
        best_run = None
        metrics = [c for c in sliced_runs.columns if c.startswith('metrics.')]
        for idx, row in sliced_runs.iterrows():
            for metric in metrics:
                if self.apply_hierarchical:
                    exec(f'{metric.split(".")[1].replace(" ","_").lower().split(" ")[0]}=row[metric]')
                else:
                    exec(f'{metric.split(".")[1].replace(" ","_").lower()}=row[metric]')

            score = eval(decision_function)
            if score < min:
                min = score
                best_run = row

        # Load model
        logger.info(message = "Retrieving best Model.")

        check_not_isinstance(obj=best_run, data_type=pd.Series, func=get_module_and_function(), attached_message='Something went wrong. The retrieved best run inside mlflow is not a pandas Series')
        model_uri = f"runs:/{best_run['run_id']}/{self.model_name}.pkl"
        self.Model = mlflow.sklearn.load_model(model_uri)

        if self.apply_hierarchical:
            folder_best_reconcile = 'best_reconciler'
            best_reconciler = os.listdir(os.path.join(best_run.artifact_uri, folder_best_reconcile))[0]
            reconciler_uri = f"runs:/{best_run['run_id']}/{folder_best_reconcile}/{best_reconciler}"
            self.best_reconciler = mlflow.sklearn.load_model(reconciler_uri)

            path_artifacts = os.path.join(best_run.artifact_uri, 'dataframes')
            with open(os.path.join(path_artifacts, 'tags.json'), 'r') as f:
                tmp = json.load(f)

            valid_true = pd.read_parquet(os.path.join(path_artifacts, 'valid_real.parquet'))
            valid_pred = pd.read_parquet(os.path.join(path_artifacts, 'valid_pred.parquet'))
            test_true  = pd.read_parquet(os.path.join(path_artifacts, 'test_real.parquet'))
            test_pred  = pd.read_parquet(os.path.join(path_artifacts, 'test_pred.parquet'))

            self.X_true = pd.concat([valid_true, test_true])
            self.X_pred = pd.concat([valid_pred, test_pred])

            self.S = pd.read_parquet(os.path.join(path_artifacts, 'summing_matrix.parquet'))
            self.tags = {k: np.array(v,dtype=object) for k,v in tmp.items()}


    def make_future_dataset(self, start_date: str, num_forecast_periods: str):
        """
        Generates a dataframe on which the predictions are applied starting from a given date

        Parameters
        ----------
        start_date : str
            The starting date
        num_forecast_periods : str
            The number of forecast periods to generate

        Returns
        -------
        pd.DataFrame
            The dataframe with future values
        """

        logger.info(message = "Generating Future data.")

        dates = pd.date_range(start = start_date, periods = num_forecast_periods, freq = self.time_granularity)
        id_preds = list(self.Model.models_dict.keys())

        future_df = pd.DataFrame(columns=['ds', 'unique_id'])
        for id_pred in id_preds:
            tmp = pd.DataFrame({"ds": dates, 'unique_id':[id_pred]*num_forecast_periods})
            future_df = pd.concat([future_df, tmp])

        self.future_df = future_df.set_index('unique_id')
    
    
    def predict(self, start_date: str, num_forecast_periods: str, decision_function: str ='rmse*0.5 + mae*0.5') -> pd.DataFrame:
        """
        Method for generate future predictions

        Parameters
        ----------
        start_date : str
            The starting date
        num_forecast_periods : str
            The number of forecast periods to generate
        decision_function : str
            Decision function to minimize in order to find the best parameters, by default 'rmse*0.5 + mape*0.5'

            The Decision function can use ONLY the evaluation parameters, beware that if an evaluation parameter
            has a space " " in its name, inside the evaluation it MUST be written as a "_", for example if using
            Percentage Coverage instead of the rmse inside the default function, this becomes 'percentage_coverage*0.5 + mape*0.5'
        
        Returns
        -------
        pd.DataFrame
            Output dataframe
        """

        logger.info(message = "START Predicting Future data.")

        self.find_best_run(decision_function=decision_function)
        self.make_future_dataset(start_date=start_date, num_forecast_periods=num_forecast_periods)
        forecast_model = self.Model.predict(self.future_df)

        if self.apply_hierarchical:
            try:
                reconcile = f'{self.best_reconciler.__class__.__name__}(method="{self.best_reconciler.method}")'
            except AttributeError:
                reconcile = f'{self.best_reconciler.__class__.__name__}()'
            
            logger.info(message = f'Applying Hierarchical Reconciliation Model: {reconcile}')

            HRM = HierarchicalReconcileModel(model_name=self.model_name, S=self.S, tags=self.tags, reconciler=reconcile)
            forecast_rec = HRM.fit_predict(X_true=self.X_true, X_pred=self.X_pred, X_future_pred=forecast_model, as_output=True)

            return forecast_model, forecast_rec
        else:
            return forecast_model, None


from .exceptions import InputTypeError, ConversionError
from typing import Any, Iterable
import pandas as pd

def check_not_isinstance(obj: Any, data_type: Any, func: str, attached_message: str = ''):
    """
    The function check if an object is of a given data type

    Parameters
    ----------
    obj : Any
        The object that needs to be checked
    data_type : Any
        The reference data type 
    func : str
        String referring to the function which is raising the errors
    attached_message : str, Optional
        Additional message to attach to the error message, by default ''

    Raises
    ------
    InputTypeError
        If the object is not of the data type specified
    """
    if not isinstance(obj, data_type):
        m = f"Type input obj {type(obj)} must be {data_type}"
        if attached_message:
            m+=f' - {attached_message}'

        raise InputTypeError(func, message=m)

def check_not_in_iterable(obj: Any, iterable: Iterable, func: str, attached_message: str = ''):
    """
    The function check if an object is in an iterable

    Parameters
    ----------
    obj : Any
        The object that needs to be checked
    iterable : Iterable
        The reference iterable
    func : str
        String referring to the function which is raising the errors
    attached_message : str, Optional
        Additional message to attach to the error message, by default ''

    Raises
    ------
    InputTypeError
        If the object is not in the iterable specified
    """
    if obj not in iterable:
        m = f"Invalid value for {obj}. Value not found in {iterable}"
        if attached_message:
            m+=f' - {attached_message}'
            
        raise InputTypeError(func, message=f"Invalid value for {obj}. Value not found in {iterable} - {attached_message}")
    
def check_datatype_convertible(obj: Any, data_type_conversion: Any, func: str, attached_message: str = ''):
    """The function checks if an object can be casted to a given datatype

    Parameters
    ----------
    obj : Any
        The object that needs to be checked
    data_type_conversion : Any
        The data type convertion function
    func : str
        String referring to the function which is raising the errors
    attached_message : str, optional
        Additional message to attach to the error message, by default ''

    Raises
    ------
    ConversionError
        If the object can not be casted
    """
    try:
        data_type_conversion(obj)
    except (ValueError, TypeError):
        m = f'The object {obj} cannot be converted into data type {data_type_conversion}'
        if attached_message:
            m += f' - {attached_message}'
        raise ConversionError(func = func, message = m)


def check_dates_complete(date_series, freq = 'D'):
    """
    Check if the dates in the given pandas.Series are complete, considering the specified frequency.

    Parameters:
        date_series (pandas.Series): A pandas.Series containing the dates.
        freq (str): The frequency level to consider for completeness check (default is 'D' for daily).

    Returns:
        bool: True if the dates are complete (no missing dates), False otherwise.
    """
    # Convert the dates in the 'date_series' to a pandas.DatetimeIndex
    date_index = pd.DatetimeIndex(date_series)

    # Calculate the expected number of dates based on the frequency
    expected_dates = pd.date_range(start=date_index.min(), end=date_index.max(), freq=freq)

    # Compare the number of unique dates in the 'date_series' with the number of expected dates
    return date_index.nunique() == len(expected_dates)


from typing import Tuple, Dict, List
import pandas as pd
from ..utility.resources import log_exception, logger
from .uc_preproc import TCPreprocess, TVPreprocess, TMPreprocess, TAPreprocess
from .model_preproc import ProphetPreprocess


# Since this class wraps other classes based on if-else statements, the typehint could not coincide with the real output type
# When we implement the other trasnformations for the other usecases, we need to check it
@log_exception(logger)
class PreprocessingClass():
    def __init__(self, usecase: str, model: str, **kwargs):
        """ Pass all the parameters that can be set as explicit keyword
        arguments.
        
        Parameters
        ----------
        usecase : str
            Parameter identifying usecase into account
        model : str
            Parameter identifying model to use
            
        """
        if usecase.lower() == 'teleconsulto':
            self.uc_preproc = TCPreprocess(hierarchy=kwargs['hierarchy'], 
                                           conversion=kwargs['conversion'], 
                                           target_col=kwargs['target_col'])
        elif usecase.lower() == 'televisita':
            self.uc_preproc = TVPreprocess(hierarchy=kwargs['hierarchy'], 
                                           conversion=kwargs['conversion'])
        elif usecase.lower() == 'telemonitoraggio':
            self.uc_preproc = TMPreprocess(hierarchy=kwargs['hierarchy'], 
                                           conversion=kwargs['conversion'])
        elif usecase.lower() == 'teleassistenza':
            self.uc_preproc = TAPreprocess(hierarchy=kwargs['hierarchy'], 
                                           conversion=kwargs['conversion'])
        else:
            raise Exception(f'Usecase "{usecase}" not managed by the Preprocessing procedures.')
        

        if model.lower() == 'prophet':
            self.model_preproc = ProphetPreprocess(hierarchy=kwargs['hierarchy'],
                                                   time_granularity=kwargs['time_granularity'],
                                                   date_col=kwargs['date_col'],
                                                   missing_data_strategy=kwargs['missing_data_strategy'])
        else:
            raise Exception(f'Model "{model}" not managed by the Preprocessing procedures.')
        
        logger.info(message = f'Instantiated the Preprocessing class for usecase: {usecase} and model: {model}.')
    
    def fit(self):
        """
        Fit method of PreprocessingClass
        """
        self.uc_preproc.fit()
        self.model_preproc.fit()

    def transform(self, X: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]:
        """
        Transform method which applies the pipeline of transformations for UC and Model 

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe to transform

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]
            Tuple containing (X, S, tags)
            - X: The processed dataframe after preprocessing
            - S: Dataframe of binary values resulting from the aggregation (incidence matrix of undirected graph)
            - tags: Dictionary containing the information about the hierarchies
        """
        X_uc = self.uc_preproc.transform(X)
        X_model = self.model_preproc.transform(X_uc)

        return X_model

    def fit_transform(self, X: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]:
        """
        fit_transform method which applies the pipeline of transformations for UC and Model 

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe to transform

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]
            Tuple containing (X, S, tags)
            - X: The processed dataframe after preprocessing
            - S: Dataframe of binary values resulting from the aggregation (incidence matrix of undirected graph)
            - tags: Dictionary containing the information about the hierarchies
        """
        X_uc = self.uc_preproc.fit().transform(X)
        X_model = self.model_preproc.fit().transform(X_uc)
        
        return X_model


import pandas as pd
from tabulate import tabulate

class InputTypeError(Exception):
    """Exception raised for errors in the function inputs."""

    def __init__(self, func : str, message : str = 'Wrong input types have been given to the function.'):
        """
        Parameters
        ----------
        func : str
            Function which is throwing the error
        message : str, optional
            Message to show 
        """
        self.message = f'{func} encountered an error: {message}'
        super().__init__(self.message)

class ColumnNotFound(Exception):
    """Exception raised when a column is not found inside the data"""
    def __init__(self, func : str, column : str, data : pd.DataFrame, message : str = 'The following column is not in the data columns list -> '):
        """
        Parameters
        ----------
        func : str
            Function which is throwing the error
        column : str
            Column not found in the data columns list
        data : pd.DataFrame
            Pandas DataFrame that thrown the error
        message : str, optional
            Message to show 
        """
        self.message = f"{func} encountered an error: {message}{column}\n\n{tabulate(data.head(4), headers='keys', tablefmt='psql')}"
        super().__init__(self.message)

class ConversionError(Exception):
    """Exception raised when two objects have not the same values"""
    def __init__(self, func : str, message : str = 'The object cannot be converted'):
        """
        Parameters
        ----------
        func : str
            Function which is throwing the error
        message : str, optional
            Message to show 
        """
        self.message = f'{func} encountered an error: {message}'
        super().__init__(self.message)


from typing import Dict, Any, List, Tuple, Union
import tomli
import os
import pandas as pd

from .exceptions import InputTypeError
from .check_utils import check_not_in_iterable, check_not_isinstance, check_datatype_convertible
from .resources import get_module_and_function, log_exception, logger


def get_configuration(config_path : str, config_file: str) -> Dict[str, Any]:
    """Read the configurations set in the toml file in the config folder and returns them
    as a dictionary

    Parameters
    ----------
    config_path : str
        Specifies the config path where the .toml configurations resides
    config_file : str
        Specifies the config file to read

    Returns
    -------
    Dict[str, Any]
        Dictionary containing the config file

    Raises
    ------
    Exception
        Raise error if .conf file is not found
    """

    if config_file not in os.listdir(config_path):
        raise Exception(f"File {config_file} not found in {config_path} please double check that the filename is correct.")

    config = tomli.load(open(os.path.join(config_path, config_file), "rb"))
    return config

def config_reading(config_dict: Dict[str, str], config_dict_parsing: Dict[str, str]) -> Tuple[dict, dict]:
    """
    Function which reads the two configuration files

    Parameters
    ----------
    config_dict : Dict[str, str]
        Dictionary of the configuration file {'config_path': path to the config file, 'config_file': config filename}
    config_dict_parsing : Dict[str, str]
        Dictionary of the REFERENCE configuration file {'config_path': path to the config file, 'config_file': config filename}

    Returns
    -------
    Tuple[dict, dict]
        Two dictionaries representing the two configuration files
    """
    config_file = get_configuration(config_dict['config_path'], config_dict['config_file'])
    config_file_parsing = get_configuration(os.path.join(config_dict_parsing['config_path'], 'config_check'), config_dict_parsing['config_file'])
    return config_file, config_file_parsing

def config_key_parsing(config_file: Dict[str, Any], config_file_parsing: Dict[str, Any]):
    """
    The function checks if the config file and the REFERENCE config file have the same keys

    Parameters
    ----------
    config_file : Dict[str, Any]
        Dictionary of the config file
    config_file_parsing : Dict[str, Any]
        Dictionary of the REFERENCE config file
    """
    for key in config_file.keys():
        check_not_in_iterable(obj = key, iterable = config_file_parsing.keys(), func = get_module_and_function())

def check_usecases(usecase: str, possible_usecases: List[str]):
    """
    Check if the usecase is in the possible usecase

    Parameters
    ----------
    usecase : str
        The usecase to check
    possible_usecases : List[str]
        The possible usecases
    """
    check_not_isinstance(obj =  usecase, data_type = str, func = get_module_and_function())
    check_not_isinstance(obj =  possible_usecases, data_type = list, func = get_module_and_function())
    check_not_in_iterable(obj = usecase, iterable = possible_usecases, func = get_module_and_function())

def check_model(model_name: str, reference_model_name: str, apply_hierarchical: bool, reference_apply_hierarchical: List[bool]):
    """
    Chek model_name and apply_hierarchical constistency

    Parameters
    ----------
    model_name : str
        Model name to check
    reference_model_name : str
        Reference model name which MUST be equal to model_name
    apply_hierarchical : bool
        Boolean value which specifies if apply hierarchical or not
    reference_apply_hierarchical : List[bool]
        List of possible values for apply_hierarchical
    """
    check_not_isinstance(obj =  model_name, data_type = str, func = get_module_and_function())
    check_not_isinstance(obj =  reference_model_name, data_type = str, func = get_module_and_function())
    check_not_isinstance(obj =  apply_hierarchical, data_type = bool, func = get_module_and_function())
    check_not_in_iterable(obj = apply_hierarchical, iterable = reference_apply_hierarchical, func = get_module_and_function())
    if model_name != reference_model_name:
        raise InputTypeError(func = get_module_and_function(), 
                             message=f"model_name {model_name} MUST be equal to reference_model_name {reference_model_name}")

def check_hyperparameters(hyperparameters: List[str], possible_hyperparameters: List[str]):
    """
    The function checks if the hyperparameters specified are allowed

    Parameters
    ----------
    hyperparameters : List[str]
        The hyperparameters specified
    possible_hyperparameters : List[str]
        The allowed hyperparameters
    """
    check_not_isinstance(obj = hyperparameters, data_type = list, func = get_module_and_function())
    check_not_isinstance(obj = possible_hyperparameters, data_type = list, func = get_module_and_function()) 
    for hyperparam in hyperparameters:
        check_not_isinstance(obj = hyperparam, data_type = str, func = get_module_and_function())
        check_not_in_iterable(obj = hyperparam, iterable = possible_hyperparameters, func = get_module_and_function())

def check_hyperparameters_value(hyperparameters: Dict[str, Union[List[str], Dict[str, Union[int, float]]]]):
    """
    The function checks if the hyperparameters are Lists or Dicts

    Parameters
    ----------
    hyperparameters : Dict[str, Union[List[str], Dict[str, Union[int, float]]]]
        The hyperparameters dict specified in the config file

    Raises
    ------
    InputTypeError
        If the value of the hyperparameters are neither Lists nor Dicts
    """
    for hyperparam_val in hyperparameters.values():
        if not isinstance(hyperparam_val, dict) and not isinstance(hyperparam_val, list):
            raise InputTypeError(func = get_module_and_function(), 
                                 message='The hyperparameters to specify must be List or Dict')

def check_missing_data_strategy(missing_data_strategy:  Union[str,int,dict], possible_strategies: List[str], 
                                possible_dictionary_keys: List[str], possible_interpolation_methods: List[str]):
    """
    The function checks the missing data strategy

    Parameters
    ----------
    missing_data_strategy : Union[str,int,dict]
        Missing data strategy to check
    possible_strategies : List[str]
        Possible strategies if the missing_data_strategy is a string
    possible_dictionary_keys : List[str]
        Possible keys if the missing_data_strategy is a dict
    possible_interpolation_methods : List[str]
        Possible interpolation methods

    Raises
    ------
    TypeError
        If missing_data_strategy is not Union[str,int,dict]
    """
    if isinstance(missing_data_strategy, str):    
        check_not_in_iterable(obj = missing_data_strategy, iterable = possible_strategies, func = get_module_and_function())
    elif isinstance(missing_data_strategy, int):
        pass
        if missing_data_strategy != 0:
            raise TypeError('Invalid value for missing_data_strategy. It must be equal to 0')
    elif isinstance(missing_data_strategy, dict):
        for key, _ in missing_data_strategy.items():
            check_not_in_iterable(obj = key, iterable = possible_dictionary_keys, func = get_module_and_function())
        check_not_in_iterable(obj = missing_data_strategy['interpolation'], iterable = possible_interpolation_methods, func = get_module_and_function())
        check_not_isinstance(obj = missing_data_strategy['order'], data_type= int, func = get_module_and_function())
    else:
        raise TypeError("Invalid type for 'missing_data_strategy' in 'configuration_utils.check_missing_data_strategy'. It must be str or int or dict")

def check_time_granularity(time_granularity: str, possible_time_granularities: List[str]):
    """
    The function checks if time_granularity is in the possible_time_granularities

    Parameters
    ----------
    time_granularity : str
        Time granularity
    possible_time_granularities : List[str]
        List of possible time granularities
    """
    check_not_isinstance(obj = time_granularity, data_type = str, func = get_module_and_function())
    check_not_isinstance(obj = possible_time_granularities, data_type = list, func = get_module_and_function()) 
    check_not_in_iterable(obj = time_granularity, iterable = possible_time_granularities, func = get_module_and_function())

def check_preprocessing(target_col: str, date_col: str):
    """
    Check if the target_col and the date_col are strings

    Parameters
    ----------
    target_col : str
        target_col in the configuration file
    date_col : str
        date_col in the configuration file
    """
    check_not_isinstance(obj = target_col, data_type = str, func= get_module_and_function())
    check_not_isinstance(obj = date_col, data_type = str, func= get_module_and_function())

def check_split_forecasting(time_granularity: str, validation_obs: int, test_obs: int,
                            num_forecast_periods: int, start_date: str):
    """
    The function checks the consistency of the split (Train, Validation, Test) and the num_forecast_periods

    Parameters
    ----------
    time_granularity : str
        Time_granularity
    validation_obs : int
        Number of observations in the validation set
    test_obs : int
        Number of observations in the test set
    num_forecast_periods : int
        Number of observations for the future
    start_date : str
        String representing the start date for future forecasting
    """
    check_not_isinstance(obj = time_granularity, data_type = str, func = get_module_and_function())
    check_not_isinstance(obj = validation_obs, data_type = int, func = get_module_and_function())
    check_not_isinstance(obj = test_obs, data_type = int, func = get_module_and_function())
    check_not_isinstance(obj = num_forecast_periods, data_type = int, func = get_module_and_function())
    check_not_isinstance(obj = start_date, data_type = str, func = get_module_and_function())
    check_datatype_convertible(obj = start_date, data_type_conversion = pd.to_datetime, func = get_module_and_function())
    
    # Check val - test - future consistency
    if validation_obs == 0 or test_obs == 0 or num_forecast_periods == 0:
        raise InputTypeError(func = get_module_and_function(), 
                             message=f"Validation size {validation_obs}, Test size {test_obs} and Future forecast periods {num_forecast_periods} MUST be greater than 0.")
    
    if test_obs >= validation_obs or num_forecast_periods >= validation_obs:
        raise InputTypeError(func = get_module_and_function(), 
                             message=f'Validation size {validation_obs} MUST be greater than Test size {test_obs} and Future forecast periods {num_forecast_periods}')

    # Check proportion consistency val - test - future depending on the time granularity
    if time_granularity == 'M':
        difference = 3
    if time_granularity == 'W': 
        difference = 4
    if time_granularity == 'D':
        difference = 7
    if time_granularity == 'H':
        difference = 24
    if (validation_obs - test_obs) < difference or (validation_obs - num_forecast_periods) < difference:
        raise InputTypeError(func = get_module_and_function(), 
                             message=f'Validation size {validation_obs} MUST be greater than Test size {test_obs} and Future forecast periods {num_forecast_periods} ' + 
                             f'of at least {difference} for time granularity {time_granularity}')

def check_hierarchy(hierarchy: Dict[str, str]):
    """
    The function checks if the values of the hierarhcy are strings

    Parameters
    ----------
    hierarchy : Dict[str, str]
        The heirarchy dictionary from the configuration file
    """
    for hier_val in hierarchy.values():
        check_not_isinstance(obj = hier_val, data_type = str, func= get_module_and_function())

def check_conversion(conversion: Dict[str, str]):
    """
        The function checks if the values of the conversion are strings

    Parameters
    ----------
    conversion : Dict[str, str]
        The conversion dictionary from the configuration file
    """
    for conversion_val in conversion.values():
        check_not_isinstance(obj = conversion_val, data_type = str, func= get_module_and_function())

# PARSING THE CONFIGURATION FILE
@log_exception(logger)
def config_parsing(config_path: str, 
                   config_filename: str, config_filename_parsing: str):
    """
    The function aggregates all the checks specified in the functions above

    Parameters
    ----------
    config_path : str
        The path to the folder which contains the configuration files
    config_filename : str
        The filename of the configuration file
    config_filename_parsing : str
        The filename of the REFERENCE configuration file
    """
    check_not_isinstance(obj = config_path, data_type = str, func = get_module_and_function)
    check_not_isinstance(obj = config_filename, data_type = str, func = get_module_and_function)
    check_not_isinstance(obj = config_filename_parsing, data_type = str, func = get_module_and_function)

    config_file, config_file_parsing = config_reading(config_dict = {'config_path': config_path, 'config_file': config_filename}, 
                                                      config_dict_parsing = {'config_path': config_path, 'config_file': config_filename_parsing})  
    config_key_parsing(config_file = config_file, config_file_parsing = config_file_parsing)
    logger.info(message = 'Check Config Sections - OK')
    check_usecases(usecase = config_file['usecase']['usecase'], possible_usecases = config_file_parsing['usecase']['usecase'])
    logger.info(message = 'Check Config Section <usecase> - OK')
    check_model(model_name = config_file['model']['model_name'], reference_model_name = config_file_parsing['model']['model_name'], 
                apply_hierarchical = config_file['model']['apply_hierarchical'], reference_apply_hierarchical = config_file_parsing['model']['apply_hierarchical'] )
    logger.info(message = 'Check Config Section <model> - OK')
    check_hyperparameters(hyperparameters = list(config_file['hyperparameters'].keys()), possible_hyperparameters = config_file_parsing['hyperparameters']['possible_hyperparameters'])
    check_hyperparameters_value(hyperparameters = config_file['hyperparameters'])
    logger.info(message = 'Check Config Section <hyperparameters> - OK')
    check_missing_data_strategy(missing_data_strategy = config_file['preprocessing']['missing_data_strategy'], possible_strategies = config_file_parsing['preprocessing']['missing_data_strategy_strings'],
                                possible_dictionary_keys = config_file_parsing['preprocessing']['missing_data_strategy_dict_keys'], possible_interpolation_methods = config_file_parsing['preprocessing']['allowed_interpolation_methods'])
    check_time_granularity(time_granularity = config_file['preprocessing']['time_granularity'], possible_time_granularities = config_file_parsing['preprocessing']['time_granularity'])
    check_preprocessing(target_col = config_file['preprocessing']['target_col'], date_col = config_file['preprocessing']['date_col'])
    logger.info(message = 'Check Config Section <Preprocessing> - OK')
    check_split_forecasting(time_granularity = config_file['preprocessing']['time_granularity'], validation_obs = config_file['validation_test_split']['validation_obs'],
                            test_obs = config_file['validation_test_split']['test_obs'], num_forecast_periods = config_file['forecasting']['num_forecast_periods'],
                            start_date =  config_file['forecasting']['start_date'])
    logger.info(message = 'Check Config Section <Forecasting - Preprocessig> - OK')


import numpy as np
import pandas as pd

from statsmodels.graphics.gofplots import qqplot

import plotly
import plotly.express as px
import plotly.graph_objects as go

def obtain_dataset(df: pd.DataFrame, gerarchia: str, id_pred: str):

    """generate list of prediction dataframe for gerarchy specified

    df : pd.DataFrame
        prediction dataframe
    gerarchia : str
        hierarchy selected
    id_pred : str
        name of id pred column

    Returns
    -------
    _type_
        _description_
    """

    df.index = [i for i in range(int(df.shape[0]/df[id_pred].unique().shape[0]))] * df[id_pred].unique().shape[0]

    if gerarchia == "Italia":
        n_level = len(gerarchia.split("/")) - 1
        split = np.array_split(df,df[id_pred].unique().shape[0])
        result = [i for i in split if i[id_pred].unique()[0].count("/") in (n_level,n_level+1)]
    else:
        n_level = len(gerarchia.split("/")) 
        split = np.array_split(df,df[id_pred].unique().shape[0])
        
        select = [i for i in split if i[id_pred].unique()[0].count("/") <= n_level]
        result = [i for i in select if gerarchia in i[id_pred][0]]
    
    return result

class plotPrediction():

    """class for plot prediction,result of job
    """

    def __init__(self, id_pred: str, data: str, pred_mean: str, pi_lower: str, pi_upper: str, use_case: str):
        
        self.id_pred = id_pred
        self.data = data
        self.pred_mean = pred_mean
        self.pi_lower = pi_lower
        self.pi_upper = pi_upper
        self.use_case = use_case

    def plot_pred(self, df_pred: pd.DataFrame, id_pred: str = None,
                  df_real: pd.DataFrame = None, id_pred_real: str = None, data_real: str = None, target: str = None, confidence_interval: bool = False) -> plotly.graph_objs._figure.Figure:
            
        """Plot prediction

        Parameters
        ----------
        df_pred : pd:DataFrame
            dataframe of prediction
        id_pred : str 
            id of prediction to be considered
        df_real : pd.DataFrame
            dataframe of real data
        target : str
            name column of real values

        Results
        -------
        plotly.graph_objs._figure.Figure
            plot prediction
        """
            
        if id_pred:
            lista_id = [id_pred]
        else:
            lista_id = df_pred[self.id_pred].unique()

        lista_fig = []
        for id_pre in lista_id:
            regione = id_pre

            df_pred_parz = df_pred[df_pred[self.id_pred] == id_pre]

            if not isinstance(df_real,type(None)):
                df_real_parz = df_real[df_real[id_pred_real] == id_pre]
                real = [go.Scatter(
                        x=df_real_parz[data_real],
                        y=df_real_parz[target],
                        marker=dict(color="blue"),
                        mode="lines",
                        name="past value (true)",showlegend=True
                    )]
            else:
                real = []

            fig = px.line(
                df_pred_parz,
                x=self.data,
                y=self.pred_mean,
                color_discrete_sequence=["green"],
                labels={self.data: "", self.pred_mean: f"numero {self.use_case}"},
                title=f"Predizione {self.use_case} - {regione}",
                width=800,
                height=500,
                template=dict(
                    layout=go.Layout(
                    title_font=dict(family="Rockwell", size=24, color="gray"),
                    title_x=0,
                    font=dict(color="green"),
                    )
                ),
            )

            fig.data[0].name = "predict value"
            fig.update_traces(showlegend=True)

            if confidence_interval:
                fig.add_traces(
                    [
                        go.Scatter(
                            x=df_pred_parz[self.data],
                            y=(df_pred_parz[self.pi_lower]),
                            mode="lines",
                            line_color="rgba(0,0,0,0)",
                            showlegend=False,
                        ),
                        go.Scatter(
                            x=df_pred_parz[self.data],
                            y=(df_pred_parz[self.pi_upper]),
                            mode="lines",
                            line_color="rgba(0,0,0,0)",
                            name="95% confidence interval",
                            fill="tonexty",
                            fillcolor="rgba(124, 252, 0, 0.2)",
                            showlegend=True,
                        )
                    ] 
                )

            fig.add_traces(real)

            fig.update_layout(legend=dict(yanchor="top", y=0.99, xanchor="left", x=0.01))

            lista_fig += [fig]

        return lista_fig

    def plot_hierarchical(self, df_pred: pd.DataFrame, gerarchia: str):

        """Plot a set of hierarchical time series

        Parameters
        ----------
        lista_df_pred : list
            list of dataframe of prediction for the low level of hierarchy
    
        Returns
        -------
        plotly.graph_objs._figure.Figure
            plot a set hierarchical time series
        """

        lista_df_pred = obtain_dataset(df_pred,gerarchia,self.id_pred)

        fig = go.Figure(
            layout=dict(
                template="plotly_white",
                title_font=dict(family="Rockwell", size=24, color="gray"),
                font=dict(color="green"),
                title=dict(text=f"Predizione gerarchia {self.use_case}"),
                legend=dict(yanchor="bottom", y=0, xanchor="left", x=-0.5),
                yaxis_title=f"numero {self.use_case}",
            )
        )

        list_fig = [
            go.Scatter(
                x=pred[self.data],
                y=pred[self.pred_mean],
                mode="lines",
                name=pred[self.id_pred][0],
                showlegend=True,
            )
            for pred in lista_df_pred
        ]

        # list_fig = [go.Scatter(x=pred[data],y=pred[pred_mean],marker=dict(color=col),mode="lines",name=pred['Id_pred'][0],showlegend=True)
        #             for pred, col in zip(lista_df_pred,["red","blue"])]

        fig.add_traces(list_fig)

        return fig

    def plot_sum_hierarchy(self, df_pred: pd.DataFrame, gerarchia: str):
        
        """Plot sum of hierarchical time series

        Parameters
        ----------
        df_pred : pd.DataFrame
            dataframe of high level of hierarchy
        lista_df_pred : list
            list of dataframe of prediction for the low level of hierarchy
        livello : str
            name of low level hierarchy

        Returns
        -------
        plotly.graph_objs._figure.Figure
            plot of sum hierarchical time series
        """

        lista_df_pred = obtain_dataset(df_pred,gerarchia,self.id_pred)
        
        df_pred_tot = lista_df_pred[0]
        lista_df_pred = lista_df_pred[1:]

        tot_hier_pred = pd.concat(lista_df_pred, axis=1)[self.pred_mean].sum(axis=1)

        fig = go.Figure(
            layout=dict(
                template="plotly_white",
                title_font=dict(family="Rockwell", size=24, color="gray"),
                font=dict(color="green"),
                title=dict(text=f"Predizione gerarchia aggregata {self.use_case}"),
                legend=dict(yanchor="top", y=0.99, xanchor="left", x=0.01),
                yaxis_title=f"numero {self.use_case}",
            )
        )

        fig.add_traces(
            [
                go.Scatter(
                    x=df_pred_tot[self.data],
                    y=df_pred_tot[self.pred_mean],
                    marker=dict(color="green"),
                    mode="lines",
                    name=f"Predizione {df_pred_tot[self.id_pred][0]}",
                    showlegend=True,
                ),
                go.Scatter(
                    x=df_pred_tot[self.data],
                    y=tot_hier_pred.values,
                    marker=dict(color="blue"),
                    mode="lines",
                    name=f"Somma predizioni {df_pred_tot[self.id_pred][0]}",
                    showlegend=True,
                ),
            ]
        )

        return fig
    
    def dist_residuals(self, df_pred: pd.DataFrame, df_real: pd.DataFrame, id_pred_real: str, target: str, id_pred: str = None) -> list:

        """Plots distribution of residuals

        Parameters
        ----------
        df_pred : dataframe of prediction
        df_real : pd.DataFrame
            dataframe of real data
        target : str
            name column of real values
        id_pred : str 
            id of prediction to be considered

        Returns
        -------
        list
            list of figure for distribution plot of residuals
        """

        if id_pred:
            lista_id = [id_pred]
        else:
            lista_id = df_pred[self.id_pred].unique()

        lista_fig = []
        for id_pre in lista_id:

            df_pred_parz = df_pred[df_pred[self.id_pred] == id_pre]
            df_real_parz = df_real[df_real[id_pred_real] == id_pre]

            res = df_pred_parz[self.pred_mean] - df_real_parz[target]

            fig = px.histogram(res,
                               title="Distribuzione residui",
                               width=800,
                               height=500,
                               template=dict(
                                    layout=go.Layout(
                                    title_font=dict(family="Rockwell", size=24, color="gray"),
                                    title_x=0,
                                    font=dict(color="green")
                                )))
            
            fig.update_layout(xaxis_title="residual",showlegend=False)

            lista_fig += [fig]

            return lista_fig
        
    def plot_residuals(self, df_pred: pd.DataFrame, df_real: pd.DataFrame, id_pred_real: str, target: str, id_pred: str = None) -> list:

        """Plots residuals

        Parameters
        ----------
        df_pred : dataframe of prediction
        df_real : pd.DataFrame
            dataframe of real data
        target : str
            name column of real values
        id_pred : str 
            id of prediction to be considered

        Returns
        -------
        list
            list of figure for plot of residuals
        """

        if id_pred:
            lista_id = [id_pred]
        else:
            lista_id = df_pred[self.id_pred].unique()

        lista_fig = []
        for id_pre in lista_id:
            
            df_pred_parz = df_pred[df_pred[self.id_pred] == id_pre]
            df_real_parz = df_real[df_real[id_pred_real] == id_pre]

            res = df_pred_parz[self.pred_mean]-df_real_parz[target]
            import matplotlib.pyplot as plt
            fig_matplot = plt.subplot() 
            qqplot_data = qqplot(res, line='45',ax = fig_matplot).gca().lines

            fig = go.Figure()

            fig.add_trace({
                'type': 'scatter',
                'x': qqplot_data[0].get_xdata(),
                'y': qqplot_data[0].get_ydata(),
                'mode': 'markers',
                'marker': {
                    'color': '#19d3f3'
                }
            })

            fig.add_trace({
                'type': 'scatter',
                'x': qqplot_data[1].get_xdata(),
                'y': qqplot_data[1].get_ydata(),
                'mode': 'lines',
                'line': {
                    'color': '#636efa'
                }

            })


            fig['layout'].update({
                'title': 'Quantile-Quantile Plot',
                'xaxis': {
                    'title': 'Theoritical Quantities',
                    'zeroline': False
                },
                'yaxis': {
                    'title': 'Sample Quantities'
                },
                'showlegend': False,
                'width': 800,
                'height': 700,
                "template":dict(
                                            layout=go.Layout(
                                            title_font=dict(family="Rockwell", size=24, color="gray"),
                                            title_x=0,
                                            font=dict(color="green")
                                        ))
            })
            lista_fig += [fig]

        return lista_fig




import pytest
import sys
import pandas as pd

from test_tc.algorithms.experiment import ExperimentClass
from test_tc.algorithms.models import ProphetModel

# This class wraps a model class and performs its methods
# we will just try its behaviour with 'prophet' without checking what it has been checked in test_algorithm_models.py


## test ExperimentClass __init__ ##
def sample_params():
    return {
        'seasonality_prior_scale' :  {'min' : 0.01, 'max' : 10.0, 'step' : 'None'},
        'weekly_seasonality' : ['True']
    }

class ExitException(Exception):
    pass

def my_exit(status=None):
    raise ExitException(f"SystemExit with status: {status}")

def test_prophet_model_creation():
    model = "prophet"
    params = sample_params()
    config_path = "./config/config_check"

    experiment = ExperimentClass(model, params, config_path)

    assert experiment.model_name == "prophet"
    assert experiment.config_path == config_path
    assert isinstance(experiment.Model, ProphetModel)

def test_invalid_model_creation(monkeypatch):
    model = "invalid_model"
    params = sample_params()
    config_path = "./config/config_check"

    # Patch sys.exit with a custom function
    monkeypatch.setattr(sys, "exit", my_exit)

    with pytest.raises(ExitException):
        ExperimentClass(model, params, config_path)
    

## test ExperimentClass fit ##
# In this test, since we already tested the fit method in test_algorithms_models.py
# we can test the check_consistency method
def create_data():
    unique_ids_train = ['a'] * 20 + ['a/b'] * 20 + ['a/c'] * 20 + ['a/d'] * 20
    unique_ids_val = ['a'] * 10 + ['a/b'] * 10 + ['a/c'] * 10 + ['a/d'] * 10

    # Creazione delle date con ripetizioni ogni 20 giorni per train_data e ogni 10 giorni per val_data
    ds_train = pd.date_range(start='2023-01-01', periods=20, freq='20D').repeat(4)
    ds_val = pd.date_range(start='2023-01-06', periods=10, freq='10D').repeat(4)

    y_train = list(range(1, 61))
    y_train += [0]*20
    y_val = list(range(61, 91))
    y_val += [0]*10

    # Creazione dei DataFrame train_data e val_data
    train_data = pd.DataFrame({'unique_id': unique_ids_train, 'ds': ds_train, 'y': y_train})
    val_data = pd.DataFrame({'unique_id': unique_ids_val, 'ds': ds_val, 'y': y_val})

    # Reorder columns
    train_data = train_data[['unique_id', 'ds', 'y']]
    val_data = val_data[['unique_id', 'ds', 'y']]

    train_data = train_data.set_index('unique_id')
    val_data = val_data.set_index('unique_id')
    train_data = train_data.sort_index()
    val_data = val_data.sort_index()
    return train_data, val_data

def test_experiment_fit():
    ## Test correct input for fit
    train_data, val_data = create_data()

    model = "prophet"
    params = sample_params()
    config_path = "./config/config_check"

    experiment = ExperimentClass(model, params, config_path)
    experiment.fit(train_data, val_data)

    ## Test not correct input for fit
    train_data.columns = ['Timestamp', 'y']
    with pytest.raises(Exception):
        experiment.fit(train_data, val_data)

## test ExperimentClass predict ##
# we check that the output is a dataframe since its properties are already checked in test_algorithm_models.py
def test_experiment_predict():
    train_data, val_data = create_data()

    model = "prophet"
    params = sample_params()
    config_path = "./config/config_check"

    experiment = ExperimentClass(model, params, config_path)
    experiment.fit(train_data, val_data)
    out = experiment.predict(val_data)
    assert type(out) == pd.DataFrame

## test ExperimentClass score ##
# we check the output format since its properties are already checked in test_algorithm_models.py
def test_experiment_score():
    train_data, val_data = create_data()

    model = "prophet"
    params = sample_params()
    config_path = "./config/config_check"

    experiment = ExperimentClass(model, params, config_path)
    experiment.fit(train_data, val_data)
    val_output = experiment.predict(val_data)

    scores_aggr, scores = experiment.score(val_data, val_output)
    
    assert type(scores_aggr) == pd.Series
    assert type(scores) == pd.DataFrame


## test ExperimentClass create_hyperparameters_table ##
# we check the output format since its properties are already checked in test_algorithm_models.py
def test_create_hyperparameters_table():
    ## Test correct input for fit
    train_data, val_data = create_data()

    model = "prophet"
    params = sample_params()
    config_path = "./config/config_check"

    experiment = ExperimentClass(model, params, config_path)
    experiment.fit(train_data, val_data)

    p = experiment.create_hyperparameters_table(params)

    assert type(p) == dict    
    p['a_d'] = {}
    for idx, param in p.items():
        reference_params = experiment.Model.models_best_params
        assert param == reference_params[idx.replace('_','/')]
       







import pytest
import logging
from unittest.mock import patch

from test_tc.utility.resources import get_configuration, log_exception, get_module_and_function

### get_configuration ###
@pytest.fixture
def mock_config_file(tmpdir):
    config_file = tmpdir.join("proprerties.toml")
    config_file.write("[section]\nkey = 'value'")
    return 'proprerties.toml'

def test_get_configuration_valid(tmpdir, mock_config_file):
    result = get_configuration("section", config_path=str(tmpdir), config_file=mock_config_file)
    assert isinstance(result, dict)
    assert result["key"] == "value"

def test_get_configuration_invalid_file(tmpdir):
    with pytest.raises(Exception):
        get_configuration("section", config_path=str(tmpdir), config_file="non_existent.toml")

def test_get_configuration_invalid_section(tmpdir, mock_config_file):
    with pytest.raises(Exception):
        get_configuration("non_existent_section", config_path=str(tmpdir), config_file=mock_config_file)

### log_exception ###
def test_log_exception_with_logger(caplog):
    logger = logging.getLogger("test_logger")

    @log_exception(logger)
    def my_function():
        raise Exception("Something went wrong")

    with caplog.at_level(logging.ERROR):
        with pytest.raises(SystemExit):
            my_function()

    assert "There was an exception in my_function" in caplog.text
    assert "Exception: Something went wrong" in caplog.text

def test_log_exception_without_logger(caplog):
    @log_exception()
    def my_function():
        raise Exception("Something went wrong")

    with caplog.at_level(logging.ERROR):
        with pytest.raises(SystemExit):
            my_function()

    assert "" in caplog.text
    assert "" in caplog.text

def test_log_exception_no_exception(caplog):
    logger = logging.getLogger("test_logger")

    @log_exception(logger)
    def my_function():
        return "Success"

    with caplog.at_level(logging.ERROR):
        result = my_function()

    assert result == "Success"
    assert "There was an exception in my_function" not in caplog.text


def test_module_and_function_format():
    result = get_module_and_function()
    assert "." in result, "Il risultato non è nel formato <module name>.<function>"

def test_module_and_function_matches_call():
    expected = "tests.test_resources.test_module_and_function_matches_call"
    result = get_module_and_function()
    assert result == expected, f"Il risultato '{result}' non corrisponde all'atteso '{expected}'"




import os
import subprocess
import mlflow
import pandas as pd

import sys
sys.path.append('../tc_uc4')
sys.path.append('../LIB_tc/test_tc')
from test_tc.utility.resources import get_configuration
from test_tc.datahandler.datahandler import DataHandler
from test_tc.algorithms.job import JobClass


# Models' name to test
models_name = [
    "Italia/Abruzzo", "Italia/Basilicata", "Italia/Bolzano", "Italia/Calabria", "Italia/Campania", "Italia/Emilia-Romagna", "Italia/Friuli-Venezia-Giulia",
    "Italia", "Italia/Lazio", "Italia/Liguria", "Italia/Lombardia", "Italia/Molise", "Italia/Piemonte", "Italia/Puglia", "Italia/Sardegna", "Italia/Sicilia",
    "Italia/Toscana", "Italia/Trento", "Italia/Umbria", "Italia/Valle D Aosta", "Italia/Veneto"
               ]


# Initial setting
experiment_name = "unittest"
python_cmd = 'python'
project_path = os.path.split(os.path.dirname(os.path.abspath(__file__)))[0]
config_path = os.path.join(project_path, "tests/config")
path_configs = experiment_name + "_path.toml"
model_configs = experiment_name + "_prophet.toml"


# Configuration path configs
experiment_folder = get_configuration("data_paths", config_path, path_configs)["experiment_folder"]
experiment_folder = os.path.join(project_path, experiment_folder)
predict_save_folder = get_configuration("data_paths", config_path, path_configs)["predict_save_folder"]


# Configuration model configs
model_conf = get_configuration("model", config_path, model_configs)
preprocessing  = get_configuration("preprocessing", config_path, model_configs)
forecasting    = get_configuration("forecasting", config_path, model_configs)
model_name = get_configuration("model", config_path, model_configs)["model_name"]
# ----------
time_granularity = preprocessing["time_granularity"]
apply_hierarchical = model_conf['apply_hierarchical']
start_date = forecasting["start_date"]
num_forecast_periods = forecasting["num_forecast_periods"]


def run_exp_job():
    
    ## TEST EXPERIMENT ##
    process = subprocess.Popen(f'{python_cmd} {os.path.join(project_path, "experiments", "execute_experiment.py")} -E {experiment_name}', shell=True)
    process.wait()

    mlflow.set_tracking_uri(experiment_folder)
    experiment_id = mlflow.get_experiment_by_name(experiment_name).experiment_id

    # Find best run
    run = mlflow.search_runs(experiment_ids=experiment_id).loc[0]

    # Test if trained models are present in artifacts
    model_uri = f"runs:/{run['run_id']}/{model_name}.pkl"
    model = mlflow.sklearn.load_model(model_uri)

    # Test model names
    for id_pred, m in model.models_dict.items():
        assert id_pred in models_name
        
        # Test parameters in specified range
        if m is not None:
            assert m.weekly_seasonality in [True, False]
            assert m.yearly_seasonality in [True, False]

    # Check apply_hierarchical = true
    best_reconciler_folder = 'best_reconciler'
    assert len(os.listdir(os.path.join(run.artifact_uri, best_reconciler_folder))) != 0


    ## TEST JOB ##
    process = subprocess.Popen(f'{python_cmd} {os.path.join(project_path, "jobs", "execute_job.py")} -E {experiment_name}',shell=True)
    process.wait()
    DH = DataHandler(predict_save_folder)
    predictions = DH.read("unittest.parquet", folder="")
    hier_predictions = DH.read("unittest_hierarchical.parquet", folder="")
    assert type(predictions) == pd.DataFrame
    assert type(hier_predictions) == pd.DataFrame



from typing import Dict
import pandas as pd

from .utils import map_columns, filter_target_col, column_to_int
from ..utility.resources import logger

class TCPreprocess():
    def __init__(self, hierarchy: Dict[str, str], conversion: Dict[str, str], target_col: str):
        """
        Init of the preprocessing class for TC

        Parameters
        ----------
        hierarchy : Dict[str, str]
            The hierarchy dictionary from the configuration file 
        conversion : Dict[str, str]
            The conversion dictionary from the configuration file
        target_col : str
            The target col from the configuration file
            (e.g. id_teleconsulto for removing duplicates)
        """
        self.hierarchy = hierarchy
        self.conversion = conversion
        self.target_col = target_col

    def fit(self, X: pd.DataFrame = None):
        """
        Fit method of TCPreprocess

        Parameters
        ----------
        X : pd.DataFrame, optional
            The pandas dataframe to fit, by default None

        Returns
        -------
        Self
        """
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform method which applies the pipeline of transformations for TC

        Parameters
        ----------
        X : pd.DataFrame
            The pandas dataframe to transform

        Returns
        -------
        pd.DataFrame
            The transformed dataframe
        """
        logger.info(message='Preparing Dataset for Model Specific Preprocessing.')
        X = map_columns(df=X, hierarchy=self.hierarchy, conversion=self.conversion)
        X = filter_target_col(df=X, target_col=self.target_col)
        return X
    
## Add the documentation of the following when implement them
class TVPreprocess():
    def __init__(self, hierarchy, conversion):
        """
        Init of the preprocessing class for TC

        Parameters
        ----------
        hierarchy : Dict[str, str]
            The hierarchy dictionary from the configuration file 
        conversion : Dict[str, str]
            The conversion dictionary from the configuration file
        """
        self.hierarchy = hierarchy
        self.conversion = conversion

    def fit(self, X: pd.DataFrame =None):
        """
        Fit method of TCPreprocess

        Parameters
        ----------
        X : pd.DataFrame, optional
            The pandas dataframe to fit, by default None

        Returns
        -------
        Self
        """
        return self

    def transform(self, X: pd.DataFrame):
        logger.info(message='Preparing Dataset for Model Specific Preprocessing.')
        X = column_to_int(df=X, hierarchy=self.hierarchy, conversion=self.conversion)
        X = map_columns(df=X, hierarchy=self.hierarchy, conversion=self.conversion)
        return X
        

class TMPreprocess():
    def __init__(self, hierarchy, conversion):
        """
        Init of the preprocessing class for TC

        Parameters
        ----------
        hierarchy : Dict[str, str]
            The hierarchy dictionary from the configuration file 
        conversion : Dict[str, str]
            The conversion dictionary from the configuration file
        """
        self.hierarchy = hierarchy
        self.conversion = conversion

    def fit(self, X: pd.DataFrame =None):
        """
        Fit method of TCPreprocess

        Parameters
        ----------
        X : pd.DataFrame, optional
            The pandas dataframe to fit, by default None

        Returns
        -------
        Self
        """
        return self

    def transform(self, X: pd.DataFrame):
        logger.info(message='Preparing Dataset for Model Specific Preprocessing.')
        X = map_columns(df=X, hierarchy=self.hierarchy, conversion=self.conversion)
        return X
        

class TAPreprocess():
    def __init__(self, hierarchy, conversion):
        """
        Init of the preprocessing class for TC

        Parameters
        ----------
        hierarchy : Dict[str, str]
            The hierarchy dictionary from the configuration file 
        conversion : Dict[str, str]
            The conversion dictionary from the configuration file
        """
        self.hierarchy = hierarchy
        self.conversion = conversion

    def fit(self, X: pd.DataFrame =None):
        """
        Fit method of TCPreprocess

        Parameters
        ----------
        X : pd.DataFrame, optional
            The pandas dataframe to fit, by default None

        Returns
        -------
        Self
        """
        return self

    def transform(self, X: pd.DataFrame):
        """
        Transform method which applies the pipeline of transformations for TC

        Parameters
        ----------
        X : pd.DataFrame
            The pandas dataframe to transform

        Returns
        -------
        pd.DataFrame
            The transformed dataframe
        """
        logger.info(message='Preparing Dataset for Model Specific Preprocessing.')
        X = map_columns(df=X, hierarchy=self.hierarchy, conversion=self.conversion)
        return X
 


from typing import Any, Dict, Tuple, Union, List
import pandas as pd

from .utils import datepart, divide_by_father, generate_time_series, get_hierarchy_column, get_lagged_column, missing_data_imputation, completing_calendar
from ..utility.resources import logger

class ProphetPreprocess():
    def __init__(self, hierarchy: Dict[str, str], time_granularity: str, 
                 date_col: str, missing_data_strategy: Union[str, int, Dict[str, Union[str, int]]]):
        """
        Init of the preprocessing class for the prophet model

        Parameters
        ----------
        hierarchy : Dict[str, str]
            The hierarchy dictionary from the configuration file 
        time_granularity : str
            The time granularity to aggregate <date_col>
        date_col : str
            The date col to aggregate with <time_granularity>
        missing_data_strategy : Union[str, int, Dict[str, Union[str, int]]]
            The missing data strategy specified in the configuration file
        """
        self.hierarchy = hierarchy
        self.time_granularity = time_granularity
        self.date_col = date_col
        self.missing_data_strategy = missing_data_strategy

    def fit(self):
        """
        Fit method of ProphetPreprocess

        Returns
        -------
        Self
        """
        return self
    
    def transform(self, X: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]:
        """
        Transform method which applies the pipeline of transformations for Prophet

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe to transform

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame, Dict[str, List[str]]]
            Tuple containing (X, S, tags)
            - X: The processed dataframe after preprocessing
            - S: Dataframe of binary values resulting from the aggregation (incidence matrix of undirected graph)
            - tags: Dictionary containing the information about the hierarchies
        """
        logger.info(message='Preparing Dataset for Prophet Modelling.')
        X, S, tags = generate_time_series(df = X, date_col = self.date_col, 
                                          time_granularity = self.time_granularity, hierarchy = self.hierarchy)
        X = completing_calendar(df = X, time_granularity = self.time_granularity)
        X = missing_data_imputation(missing_data_strategy = self.missing_data_strategy, df = X)

        return X, S, tags

class PreprocessBoosting():
    def __init__(self, datepart: bool = True, preds: bool = True, target: bool = True, conf_inter: bool = True,
                 lags: list = None, n_hier: int = None,
                 father_preds: bool = False, sons_preds: bool = False, father_target: bool = False,
                 sons_target: bool = False, divide_target: bool = False):

        """preprocessing class to obtain dataset input for boosting models

        Parameters
        ----------
        datepart : bool,
            if True adds date information to dataset
        preds : bool
            if True adds predictions to dataset
        target : bool,
            if True adds Target value to dataset
        conf_inter : bool,
            if True adds confidence intervals to dataset
        lags : list
            if True lags the predictions columns and adds them to the dataset
        n_hier : int
            level of hierarchy considered, if None considers all levels
        father_preds : bool
            if True adds father column of the predictions of the hierarchy level (selected by n_hier) to dataset
        sons_preds : bool
            if True adds all sons of the predictions of the hierarchy level (selected by n_hier) to dataset
        father_target : bool
            if True adds father of target of the hierarchy level (selected by n_hier) to dataset
        sons_target : bool
            if True adds all sons of target of the hierarchy level (selected by n_hier) to dataset
        divide_target : bool
            if True divide all target sons by target father
        """
        self.datepart = datepart
        self.preds = preds
        self.target = target
        self.conf_inter = conf_inter
        self.lags = lags
        self.n_hier = n_hier
        self.father_preds = father_preds
        self.sons_preds = sons_preds
        self.father_target = father_target
        self.sons_target = sons_target
        self.divide_target = divide_target

    def transform(self, df_real: pd.DataFrame, df_preds: pd.DataFrame,
                  date_real: str = "ds", date_preds: str = "Timestamp",
                  values_real: str = "y", values_preds: str = "Pred_Mean",
                  hier_real: str = "unique_id", hier_preds: str = "Id_Pred") -> pd.DataFrame:
        """Transforms a DataFrame

        Parameters
        ----------
        df_real : pd.DataFrame
            dataframe of target values
        df_preds : pd.DataFrame
            dataframe of prediction values
        date_real : str
            name of date column (target dataframe)
        date_preds : str
            name of date column (prediction dataframe)
        values_real : str
            name of value (y_true) column (target dataframe)
        values_preds : str
            name of value prediction (y_hat) column (prediction dataframe)
        hier_real : str
            name of hierarchy id column (target dataframe)
        hier_preds : str
            name of hierarchy id column (prediction dataframe)

        Returns
        -------
        list[pd.DataFrame]
            list of dataframes with extracted features for boosting models
        """

        dfs = []

        # Add target cols
        if self.target:
            target_list = get_hierarchy_column(df_real,date_real,values_real,hier_real,self.n_hier,sons=self.sons_target,
                                               father=self.father_target,type_="Target")

            if self.divide_target:
                target_list = [divide_by_father(target) for target in target_list]

            dfs.append(target_list)

        # Add predictions cols
        if self.preds:
            preds_list = get_hierarchy_column(df_preds,date_preds,values_preds,hier_preds,self.n_hier,sons=self.sons_preds,
                                            father=self.father_preds,type_="Prophet")

            dfs.append(preds_list)

            if self.lags:
                dfs.append([get_lagged_column(pro,self.lags) for pro in preds_list])

        # Add confidence interval cols
        if self.conf_inter:
            dfs.append(get_hierarchy_column(df_preds,date_preds,"Pi_Lower_95",hier_preds,self.n_hier,sons=self.sons_preds,
                                            father=self.father_preds,type_="Pi_Lower"))
            dfs.append(get_hierarchy_column(df_preds,date_preds,"Pi_Upper_95",hier_preds,self.n_hier,sons=self.sons_preds,
                                            father=self.father_preds,type_="Pi_Upper"))

        dfs = [pd.concat([i[j] for i in dfs],axis=1) for j in range(0,len(dfs[0]))]


        if self.datepart:
            timestamp = pd.Series(df_real[date_real].unique()).sort_values()
            date_col = datepart(timestamp)
            dfs = [pd.concat([i,date_col],axis=1) for i in dfs]

        return dfs

    @staticmethod
    def get_X_y(df: pd.DataFrame, n_hier: int) -> tuple[pd.DataFrame, pd.DataFrame]:
        target_cols = [c for c in df.columns if c.count("/") == n_hier - 1 and "Target" in c]
        feature_cols = [c for c in df.columns if "Target" not in c]

        X_train = df[feature_cols]
        y_train = df[target_cols]
        return X_train, y_train



import os
from typing import Any

import geopandas as gpd
import pandas as pd
from typing_extensions import Self

from ..utility.exceptions import InputTypeError
from ..utility.resources import log_exception, logger

class DataHandler:
    def __init__(self, data_folder : str) -> Self:
        """Class which will handle the read and write of .parquet data

        Parameters
        ----------
        data_folder : str
            Main folder where to find the data, if given as a parameter
        """
        if type(data_folder) != str:
            raise InputTypeError('DataHandler')

        self.data_folder = data_folder

    @log_exception(logger)
    def read(self, filename: str, folder: str = "input", in_attr: bool = False, **kwargs: Any) -> pd.DataFrame:
        """
        Read datafile given a folder from <data_folder>/<folder> and the filename. Supported extensions are
        csv and parquet. See: https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html
        and https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html for more info
        regarding the additional parameters

        Parameters
        ----------
        filename : str
            File name to read
        folder : str, optional
            Folder where the file resides starting from <data_folder>, the path to the default folder will be
            /../../data/<folder>, by default 'input'
        in_attr : bool, optional
            Whether to save the dataset in the Class attributes or not, by default False

        Returns
        -------
        pd.DataFrame
            Pandas datafreame read

        Raises
        ------
        Exception
            Returns an error if the file estension is not supported
        """

        if type(filename) != str or type(folder) != str or type(in_attr) != bool:
            raise InputTypeError('DataHandler.read')

        datatype = os.path.splitext(filename)[-1].lower()
        supp_datatype = [".parquet", ".csv"]

        if datatype not in supp_datatype:
            raise Exception(
                f'File extension "{datatype}" not supported. Supported extensions are {supp_datatype}'
            )

        logger.info(message=f"START loading {os.path.join(self.data_folder, folder, filename)}")

        # path = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(self.data_folder, folder, filename)

        if datatype == ".csv":
            df = pd.read_csv(filepath_or_buffer=path, **kwargs)

        elif datatype == ".parquet":
            if "engine" not in kwargs.keys():
                kwargs["engine"] = "fastparquet"

            df = pd.read_parquet(path=path, **kwargs)

        logger.info(message=f"DONE loading {os.path.join(self.data_folder, folder, filename)}")
        
        if in_attr:
            self.df = df

        return df

    @log_exception(logger)
    def read_geodataframe(self, filename: str, folder: str = "input", in_attr: bool = False, **kwargs: Any) -> gpd.GeoDataFrame:
        """
        Read datafile given a folder from /../../data/<folder> and the filename. Supported extensions are
        parquet. See: https://geopandas.org/en/stable/docs/reference/api/geopandas.read_parquet.html
        for more info regarding the additional parameters

        Parameters
        ----------
        filename : str
            File name to read
        folder : str, optional
            Folder where the file resides starting from <data_folder>, the path to the default folder will be
            /../../data/<folder>, by default 'input'
        in_attr : bool, optional
            Whether to save the dataset in the Class attributes or not, by default False

        Returns
        -------
        pd.DataFrame
            Pandas datafreame read

        Raises
        ------
        Exception
            Raises an error if the file estension is not supported
        """

        if type(filename) != str or type(folder) != str or type(in_attr) != bool:
            raise InputTypeError('DataHandler.read_geodataframe')
        
        datatype = os.path.splitext(filename)[-1].lower()
        supp_datatype = [".parquet"]

        if datatype not in supp_datatype:
            raise Exception(
                f'File extension "{datatype}" not supported. Supported extensions are {supp_datatype}'
            )

        logger.info(message=f"START loading {os.path.join(self.data_folder, folder, filename)}")

        path = os.path.join(self.data_folder, folder, filename)
        gdf = gpd.read_parquet(path=path, **kwargs)

        logger.info(message=f"DONE loading {os.path.join(self.data_folder, folder, filename)}")

        if in_attr:
            self.gdf = gdf

        return gdf

    @log_exception(logger)
    def write(
        self, data: pd.DataFrame, filename: str, folder: str = "output", **kwargs: Any
    ) -> None:
        """Writes a pandas DataFrame into the selected folder. Supported extensions are
        csv and parquet. See: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_parquet.html
        and https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_csv.html for more info
        regarding the additional parameters

        Parameters
        ----------
        data : pd.DataFrame
            Pandas DataFrame to be written
        filename : str
            Filename to be writter. It must contain the correct extension, currently supported
            extensions are csv and parquet
        folder : str, optional
            Folder where the file will be written into starting from <data_folder>, the path to the default folder will be
            /../../data/<folder>, by default 'output'
        """
        if  type(data) != pd.DataFrame or type(filename) != str or type(folder) != str:
            raise InputTypeError('DataHandler.write')
        
        datatype = os.path.splitext(filename)[-1].lower()
        supp_datatype = [".parquet", ".csv"]

        if datatype not in supp_datatype:
            raise Exception(
                f'File extension "{datatype}" not supported. Supported extensions are {supp_datatype}'
            )
        if not os.path.isdir(os.path.join(self.data_folder, folder)):
            raise Exception(
                f'Can not write "{filename}" because folder "{os.path.join(self.data_folder, folder)}". Please create it.'
            )

        path = os.path.join(self.data_folder, folder, filename)

        logger.info(message=f"START writing {os.path.join(self.data_folder, folder, filename)}")

        if datatype == ".csv":
            if "infex" not in kwargs.keys():
                kwargs["index"] = False

            data.to_csv(path_or_buf=path, **kwargs)

        elif datatype == ".parquet":
            if "engine" not in kwargs.keys():
                kwargs["engine"] = "fastparquet"

            data.to_parquet(path=path, **kwargs)

        logger.info(message=f"DONE writing {os.path.join(self.data_folder, folder, filename)}")



import pytest
from test_tc.algorithms.models import *  

# --------- ProphetModel Tests --------- #
## test ProphetModel __init__ ##
# The __init__ test also the method make_grid_params 
@pytest.fixture
def sample_params():
    return {'weekly_seasonality' : ['True'],
            'yearly_seasonality' : ['True']
}

def test_prophet_model_init(sample_params):
    config_path = "./config/config_check"
    max_na_ratio = 0.3
    model = ProphetModel(sample_params, config_path, max_na_ratio)

    assert model.config_path == config_path
    assert model.max_na_ratio == max_na_ratio
    assert model.grid_params is not None
    assert model.grid_params == {'weekly_seasonality' : [True], 
                                 'yearly_seasonality' : [True],
                                 'interval_width' : [0.95]}

##  test ProphetModel fit method ##
def create_data():
    unique_ids_train = ['a'] * 20 + ['a/b'] * 20 + ['a/c'] * 20 + ['a/d'] * 20
    unique_ids_val = ['a'] * 10 + ['a/b'] * 10 + ['a/c'] * 10 + ['a/d'] * 10

    # Creazione delle date con ripetizioni ogni 20 giorni per train_data e ogni 10 giorni per val_data
    ds_train = pd.date_range(start='2023-01-01', periods=20, freq='20D').repeat(4)
    ds_val = pd.date_range(start='2023-01-06', periods=10, freq='10D').repeat(4)

    y_train = list(range(1, 61))
    y_train += [0]*20
    y_val = list(range(61, 91))
    y_val += [0]*10

    # Creazione dei DataFrame train_data e val_data
    train_data = pd.DataFrame({'unique_id': unique_ids_train, 'ds': ds_train, 'y': y_train})
    val_data = pd.DataFrame({'unique_id': unique_ids_val, 'ds': ds_val, 'y': y_val})

    # Reorder columns
    train_data = train_data[['unique_id', 'ds', 'y']]
    val_data = val_data[['unique_id', 'ds', 'y']]
    return train_data, val_data

def test_prophet_model_fit():
    config_path = "./config/config_check"
    max_na_ratio = 0.3
    params = {'weekly_seasonality' : ['True'],
            'yearly_seasonality' : ['True']}
    reference_params = {'weekly_seasonality' : True, 
                                 'yearly_seasonality' : True,
                                 'interval_width' : 0.95}
    model = ProphetModel(params, config_path, max_na_ratio)

    train_data, val_data = create_data()

    # Specificare il numero di ripetizioni desiderate
    num_repetitions = 20

    # Creazione dei DataFrame con più righe
    train_data = pd.concat([train_data] * num_repetitions)
    val_data = pd.concat([val_data] * num_repetitions)

    train_data = train_data.set_index('unique_id')
    val_data = val_data.set_index('unique_id')
    train_data = train_data.sort_index()
    val_data = val_data.sort_index()

    # Call the fit method
    model.fit(train_data, val_data)

    # Check if models_dict and models_best_params are populated
    assert model.models_dict != {}
    assert model.models_best_params != {}
    for k, best_param in model.models_best_params.items():
        if k != 'a/d':
            assert best_param == reference_params
        else:
            assert best_param == {}

## test ProphetModel refit method ##
def test_prophet_model_refit():
    # first fit
    config_path = "./config/config_check"
    max_na_ratio = 0.3
    params = {'weekly_seasonality' : ['True'],
            'yearly_seasonality' : ['True']}
    reference_params = {'weekly_seasonality' : True, 
                                 'yearly_seasonality' : True,
                                 'interval_width' : 0.95}
    model = ProphetModel(params, config_path, max_na_ratio)

    train_data, val_data = create_data()

    # Specificare il numero di ripetizioni desiderate
    num_repetitions = 20

    # Creazione dei DataFrame con più righe
    train_data = pd.concat([train_data] * num_repetitions)
    val_data = pd.concat([val_data] * num_repetitions)

    train_data = train_data.set_index('unique_id')
    val_data = val_data.set_index('unique_id')
    train_data = train_data.sort_index()
    val_data = val_data.sort_index()

    # Call the fit method
    model.fit(train_data, val_data)

    # Then refit
    model.refit(train_data)

    # Check if models_dict and models_best_params are populated
    assert model.models_dict != {}
    assert model.models_best_params != {}
    for k, best_param in model.models_best_params.items():
        if k != 'a/d':
            assert best_param == reference_params
        else:
            assert best_param == {}

## test ProphetModel predict method ##
# The function checks also the method prepare_output
def test_prophet_model_predict():
    # first fit
    config_path = "./config/config_check"
    max_na_ratio = 0.3
    params = {'weekly_seasonality' : ['True'],
            'yearly_seasonality' : ['True']}
   
    model = ProphetModel(params, config_path, max_na_ratio)

    train_data, val_data = create_data()

    # Specificare il numero di ripetizioni desiderate
    num_repetitions = 20

    # Creazione dei DataFrame con più righe
    train_data = pd.concat([train_data] * num_repetitions)
    val_data = pd.concat([val_data] * num_repetitions)

    train_data = train_data.set_index('unique_id')
    val_data = val_data.set_index('unique_id')
    train_data = train_data.sort_index()
    val_data = val_data.sort_index()

    # Call the fit method
    model.fit(train_data, val_data)

    # Then predict
    output = model.predict(val_data)

    # Assert output columns
    assert np.all(output.columns == ['Timestamp', 'Id_Pred' , 'Pred_Mean'  ,'Sigma',  'Pi_Lower_95' , 'Pi_Upper_95'])
    
    # Assert Id_Pred consistency
    assert np.all(output['Id_Pred'].unique() == ['a', 'a/b', 'a/c', 'a/d'])

    # Assert predictions for 'a/d'
    for col in ['Pred_Mean', 'Sigma', 'Pi_Lower_95', 'Pi_Upper_95']:
        assert  np.all(output[output['Id_Pred'] == 'a/d'][col] == 0)

## test ProphetModel score method ## 
def test_prophet_model_score():
    # Obtain predictions
    config_path = "./config/config_check"
    max_na_ratio = 0.3
    params = {'weekly_seasonality' : ['True'],
            'yearly_seasonality' : ['True']}
   
    model = ProphetModel(params, config_path, max_na_ratio)

    train_data, val_data = create_data()

    # Specificare il numero di ripetizioni desiderate
    num_repetitions = 20

    # Creazione dei DataFrame con più righe
    train_data = pd.concat([train_data] * num_repetitions)
    val_data = pd.concat([val_data] * num_repetitions)

    train_data = train_data.set_index('unique_id')
    val_data = val_data.set_index('unique_id')
    train_data = train_data.sort_index()
    val_data = val_data.sort_index()

    # Call the fit method
    model.fit(train_data, val_data)

    # Then predict
    df_pred = model.predict(val_data)

    # Then score
    score = model.score(val_data, df_pred)
    
    # Assert score columns
    assert np.all(score.columns == ['Id_Pred', 'MAE', 'MAPE', 'RMSE', 'MSE', 'R2', 'Percentage Coverage'])


# --------- HierarchicalReconcileMethod Tests --------- #
## test HierarchicalReconcileMethod __init__ ##
def sample_data():
    model_name = "test_model"
    S = pd.DataFrame({
        "column1": [1, 2, 3, 4, None],
        "column2": [10, 20, None, 40, 50]
    })
    tags = {"tag1": ["column1", "column2"]}
    reconciler = None
    max_na_ratio = 0.5

    return model_name, S, tags, reconciler, max_na_ratio

def test_init_with_none_reconciler():
    model_name, S, tags, reconciler, max_na_ratio = sample_data()
    model = HierarchicalReconcileModel(model_name, S, tags, reconciler, max_na_ratio)

    assert model.model_name == model_name
    assert model.S.equals(S)
    assert model.tags == tags
    assert model.max_na_ratio == max_na_ratio

def test_init_with_custom_reconciler():
    model_name, S, tags, reconciler, max_na_ratio = sample_data()
    reconciler = BottomUp()
    model = HierarchicalReconcileModel(model_name, S, tags, reconciler, max_na_ratio)

    assert model.model_name == model_name
    assert model.S.equals(S)
    assert model.tags == tags
    assert len(model.reconciler) == 1
    assert model.max_na_ratio == max_na_ratio


## test HierarchicalReconcileMethod prepare_dataframe_to_hf ##
def prepare_data():
    X_pred = pd.DataFrame({
        'Timestamp': ['2023-07-20', '2023-07-21'],
        'Pred_Mean': [10, 15],
        'Id_Pred': [1, 2]
    })
    X_true = pd.DataFrame({
        'Timestamp': ['2023-07-20', '2023-07-21'],
        'y': [8, 12]
    })

    return X_pred, X_true

def test_prepare_dataframe_to_hf_with_X_true():
    X_pred, X_true = prepare_data()
    model_name, S, tags, reconciler, max_na_ratio = sample_data()
    model = HierarchicalReconcileModel(model_name, S, tags, reconciler, max_na_ratio)

    result = model.prepare_dataframe_to_hf(X_pred, model_name, X_true)

    expected_columns = ['ds', 'Test_Model', 'y']
    assert result.columns.tolist() == expected_columns
    assert result.index.tolist() == [1, 2]
    assert result['Test_Model'].tolist() == [10, 15]
    assert result['y'].tolist() == [8, 12]

def test_prepare_dataframe_to_hf_without_X_true():
    X_pred, _ = prepare_data()
    model_name, S, tags, reconciler, max_na_ratio = sample_data()
    model = HierarchicalReconcileModel(model_name, S, tags, reconciler, max_na_ratio)

    result = model.prepare_dataframe_to_hf(X_pred, model_name)

    expected_columns = ['ds', 'Test_Model']
    assert result.columns.tolist() == expected_columns
    assert result.index.tolist() == [1, 2]
    assert result['Test_Model'].tolist() == [10, 15]


## test HierarchicalReconcileMethod fit ##
def test_HierarchicalReconcileMethod_fit():
    X_pred, X_true = prepare_data()
    model_name, S, tags, reconciler, max_na_ratio = sample_data()
    model = HierarchicalReconcileModel(model_name, S, tags, reconciler, max_na_ratio)

    updated_self = model.fit(X_true, X_pred, X_pred)

    assert updated_self.X.columns.tolist() == [ 'ds', 'Test_Model', 'y']  
    assert updated_self.X_hat.columns.tolist() == [ 'ds', 'Test_Model']  


## test HierarchicalReconcileMethod predict ##
# The test checks also the prepare_output method
from hierarchicalforecast.utils import aggregate
import warnings

def prepare_data_reconcile():
    df = pd.DataFrame(columns = ['level0',	'level1',	'ds',	'y'])
    df['ds'] = ['2023-07-20', '2023-07-21', '2023-07-22']*2
    df['y'] = 1
    df['level0'] = 'a'
    df['level1'] =  ['b', 'b', 'b', 'c', 'c', 'c']
    spec = [['level0'], ['level0', 'level1'],]

    Y_df, S_df, tags = aggregate(df=df, spec=spec)
    Y_df = Y_df.reset_index()
    return Y_df, S_df, tags 

def prepare_input_reconcile():
    X_pred = pd.DataFrame({
        'Timestamp': ['2023-07-20', '2023-07-21', '2023-07-22', '2023-07-23',
                      '2023-07-24', '2023-07-25', '2023-07-26', '2023-07-27']*3,
        'Pred_Mean': [10] * 8 + [4] * 8 + [2] * 8,
        'Id_Pred': ['a'] * 8 + ['a/b'] * 8 + ['a/c'] * 8,
        'Pi_Upper_95': [11] * 8 + [5] * 8 + [3] * 8,
        'Pi_Lower_95': [9] * 8 + [3] * 8 + [2] * 8

    })

    X_true = pd.DataFrame({
        'ds': ['2023-07-20', '2023-07-21', '2023-07-22', '2023-07-23',
                      '2023-07-24', '2023-07-25', '2023-07-26', '2023-07-27']*3,
        'y': [2] * 8 + [1] * 8 + [1] * 8,
        'unique_id':  ['a'] * 8 + ['a/b'] * 8 + ['a/c'] * 8
    }).set_index('unique_id')

    X_future_pred = pd.DataFrame({
        'Timestamp': ['2023-07-28', '2023-07-29', '2023-07-30']*3,
        'Pred_Mean': [10, 10, 10, 4, 4, 2, 5, 5, 2],
        'Id_Pred': ['a', 'a', 'a', 'a/b', 'a/b', 'a/b', 'a/c', 'a/c', 'a/c']
    })

    return X_true, X_pred, X_future_pred

def test_HierarchicalReconcileMethod_predict():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        _, S_df, tags = prepare_data_reconcile()
        X_true, X_pred, X_future_pred = prepare_input_reconcile()
        h_model = HierarchicalReconcileModel(model_name = 'test_model', S = S_df, tags = tags,
                                            reconciler = [TopDown(method = 'average_proportions')])
        h_model.fit(X_true, X_pred, X_future_pred)

        Y_rec = h_model.predict(as_output = True)

    # Assert if Y_rec is consistent with as_output = True
    assert Y_rec.columns.tolist() == ['Timestamp', 'Id_Pred', 'Pred_Mean', 'Sigma', 'Pi_Lower_95','Pi_Upper_95']
    # Assert if the reconciliation is logically correct
    for ds in Y_rec['Timestamp'].unique():
        check_df = Y_rec[Y_rec['Timestamp'] == ds].reset_index()
        assert check_df['Pred_Mean'][0] == check_df['Pred_Mean'][1:].sum()


## test HierarchicalReconcileMethod score ##
def test_HierarchicalReconcileMethod_score():
    import math
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        _, S_df, tags = prepare_data_reconcile()
        X_true, X_pred, X_future_pred = prepare_input_reconcile()
        h_model = HierarchicalReconcileModel(model_name = 'test_model', S = S_df, tags = tags,
                                            reconciler = [TopDown(method = 'average_proportions')])
        h_model.fit(X_true, X_pred, X_future_pred)

        # Check score NOT equal values
        # just check the columns
        Y_rec = h_model.predict(as_output = True)
        X_future_true = X_future_pred
        X_future_true.rename(columns = {'Id_Pred': 'unique_id'}, inplace= True)
        X_future_true = X_future_true.set_index('unique_id')
        X_future_true.columns = ['ds', 'y'] 
        _, scores = h_model.score(X_true = X_future_true.reset_index(), X_pred = Y_rec, check_model_pred = Y_rec, hierarchical_output = False)

        assert scores.columns.tolist() == ['Id_Pred', 'MAE', 'MAPE', 'RMSE', 'MSE', 'R2', 'Percentage Coverage']

        # Check score equal values
        # check the expected results
        X_future_true['y'] = list(Y_rec['Pred_Mean'])
        scores_aggr, _ = h_model.score(X_true = X_future_true.reset_index(), X_pred = Y_rec, check_model_pred = Y_rec, hierarchical_output = False)
        expected_scores = [0.0, 0.0, 0.0, 0.0, np.nan, 0.0]
        for i, (x, y) in enumerate(zip(scores_aggr.values, expected_scores)):
            if math.isnan(x):
                assert math.isnan(y)
            else:
                assert x == y

## test HierarchicalReconcileMethod best_reconciler ##
def test_HierarchicalReconcileMethod_best_reconciler():
    with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _, S_df, tags = prepare_data_reconcile()
            X_true, X_pred, X_future_pred = prepare_input_reconcile()
            h_model = HierarchicalReconcileModel(model_name = 'test_model', S = S_df, tags = tags,
                                                reconciler = [TopDown(method = 'average_proportions')])
            h_model.fit(X_true, X_pred, X_future_pred)

            X_future_true = X_future_pred.copy()
            X_future_true.rename(columns = {'Id_Pred': 'unique_id'}, inplace= True)
            X_future_true = X_future_true.set_index('unique_id')
            X_future_true.columns = ['ds', 'y']

            best_reconciler = h_model.best_reconciler(X_true=X_true, X_pred=X_pred, X_future_true = X_future_true, X_future_pred=X_future_pred)
            reconciler_string = str(best_reconciler).split(' ')[0]
            top_down_string = reconciler_string.split('.')[2]
            assert top_down_string == 'TopDown'


def test_boosting_reconciler_init():
    model = BoostingReconcileModel()
    assert isinstance(model.cat_features, list)
    assert isinstance(model.xgb_params, dict)
    assert isinstance(model.preprocess_params, dict)
    
    xgb = {"learning_rate": 0.2, "n_iterations": 4}
    preproc = {"father_target": True, "father_preds": False}
    cat_features = ["dayofweek", "month"]
    model = BoostingReconcileModel(
        params_xgb=xgb,
        params_preprocess=preproc,
        cat_features=cat_features
        )
    assert model.cat_features == cat_features
    assert model.xgb_params == xgb
    assert model.preprocess_params == preproc




